# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transformer-based machine translation model."""

import copy
import dataclasses
import typing
from typing import Any, Optional, Tuple, Type, TypeVar
from aqt.jax_legacy.jax import flax_attention as aqt_flax_attention
from aqt.jax_legacy.jax import flax_layers as aqt_flax_layers
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax.flax import struct as flax_struct
from flax import linen as nn
import jax
from jax import lax
import jax.numpy as jnp
import numpy as np

T = TypeVar('T')

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass


def hardware_bernoulli(rng_key, p=np.float32(0.5), shape=None):
  del rng_key  # unused
  return lax.rng_uniform(0.0, 1.0, shape) < p


def set_hardware_bernoulli():
  jax.random.bernoulli = hardware_bernoulli


def shift_right(x, axis=1):
  """Shift the input to the right by padding on axis 1."""
  pad_widths = [(0, 0)] * len(x.shape)
  pad_widths[axis] = (1, 0)
  padded = jnp.pad(
      x, pad_widths, mode='constant', constant_values=x.dtype.type(0))
  return padded[:, :-1]


class AddPositionEmbs(nn.Module):
  """Adds learned positional embeddings to the inputs.

  Attributes:
    max_len: maximum length of sequence during eval.
    min_timescale: minimum scale that will be applied at each position
    max_timescale: maximum scale that will be applied at each position
    decode: whether to run in decoding mode (one input fed in at a time)
  """

  max_len: int
  min_timescale: float
  max_timescale: float
  decode: bool

  @nn.compact
  def __call__(self, inputs, inputs_positions=None):
    """Adds positional embeddings to the inputs.

    Args:
      inputs: input data
      inputs_positions: input position indices for packed sequences.

    Returns:
      output: `(bs, timesteps, in_dim)`
    """
    _, sequence_length, num_channels = inputs.shape
    num_timescales = num_channels // 2
    max_timescale = self.max_timescale
    min_timescale = self.min_timescale
    max_len = self.max_len
    log_timescale_increment = (
        np.log(max_timescale / min_timescale) / (float(num_timescales) - 1))
    inv_timescales = min_timescale * np.exp(
        np.arange(num_timescales).astype('float32') * -log_timescale_increment)

    if inputs_positions is None:
      inputs_positions = np.expand_dims(
          np.arange(sequence_length).astype('float32'), 0)

    if self.decode:
      inputs_positions = np.expand_dims(np.arange(max_len).astype('float32'), 0)

    scaled_time = (
        jnp.expand_dims(inputs_positions.astype('float32'), 2) *
        jnp.expand_dims(np.expand_dims(inv_timescales, 0), 0))
    signal = jnp.concatenate(
        [jnp.sin(scaled_time), jnp.cos(scaled_time)], axis=2)
    signal = jnp.pad(
        signal, [[0, 0], [0, 0], [0, np.mod(num_channels, 2)]],
        mode='constant',
        constant_values=inputs.dtype.type(0))

    if self.decode:
      is_initialized = self.has_variable('cache', 'cache_index')
      cache_index = self.variable('cache', 'cache_index',
                                  lambda: jnp.array(0, dtype=jnp.int32))
      if is_initialized:
        i = cache_index.value
        one = jnp.array(1, jnp.int32)
        cache_index.value = i + one
        _, _, df = signal.shape
        signal = lax.dynamic_slice(signal, jnp.array((0, i, 0)), (1, 1, df))
    if self.decode:
      # just needed to set correct shape on init.
      return inputs + signal[:, :1, :]
    else:
      return inputs + signal


class MlpBlock(nn.Module):
  """Transformer MLP / feed-forward block."""

  @dataclass
  class HParams:
    dense_1: aqt_flax_layers.DenseAqt.HParams
    dense_2: aqt_flax_layers.DenseAqt.HParams

  hparams: HParams
  train: bool
  mlp_dim: int
  dynamic_context: quant_config.DynamicContext
  dtype: Any
  dropout_rate: float
  deterministic: bool
  kernel_init: Any = nn.initializers.xavier_uniform()
  bias_init: Any = nn.initializers.normal(stddev=1e-6)

  # TODO(shivaniagrawal): parametrize paxis_name instead of hardcoding.
  @nn.compact
  def __call__(
      self,
      inputs,
      *,
      padding_mask: jnp.ndarray,
  ):
    """Applies Transformer MlpBlock module."""

    batch_size, sequence_length, channel_size = inputs.shape
    inputs = inputs.reshape((batch_size * sequence_length, channel_size))
    shape_utils.assert_shapes_equal(padding_mask.shape,
                                    (batch_size, sequence_length, 1))
    padding_mask = padding_mask.reshape((batch_size * sequence_length, 1))
    x = aqt_flax_layers.DenseAqt(
        features=self.mlp_dim,
        dtype=self.dtype,
        paxis_name='batch',
        train=self.train,
        dynamic_context=self.dynamic_context,
        hparams=self.hparams.dense_1,
        kernel_init=self.kernel_init,
        bias_init=self.bias_init,
        name='dense_1')(
            inputs, padding_mask=padding_mask)

    x = nn.relu(x)
    x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=self.deterministic)

    output = aqt_flax_layers.DenseAqt(
        # We have relu before this layer, x would only contain positive values.
        features=channel_size,
        dtype=self.dtype,
        paxis_name='batch',
        train=self.train,
        dynamic_context=self.dynamic_context,
        hparams=self.hparams.dense_2,
        kernel_init=self.kernel_init,
        bias_init=self.bias_init,
        name='dense_2')(
            x, padding_mask=padding_mask)

    output = nn.Dropout(rate=self.dropout_rate)(
        output, deterministic=self.deterministic)
    output = output.reshape((batch_size, sequence_length, channel_size))
    return output


class Encoder1DBlock(nn.Module):
  """Transformer encoder layer.

  Attributes:
    qkv_dim: dimension of the query/key/value
    mlp_dim: dimension of the mlp on top of attention block
    num_heads: number of heads
    update_bounds: Bool whether to update activation bounds.
    train: Whether model is training.
    hparams: hyperparameters
    collect_acts_stats: Whether to tag activations to record statistics.
    dtype: the dtype of the computation (default: float32)
    dropout_rate: dropout rate
    attention_dropout_rate: dropout rate for attention weights
    deterministic: bool, deterministic or not (to apply dropout)
  """

  @dataclass
  class HParams:
    mlp_block: MlpBlock.HParams
    attention: aqt_flax_attention.MultiHeadDotProductAttentionAqt.HParams
    layer_norm: aqt_flax_layers.LayerNormAqt.HParams

  hparams: HParams
  qkv_dim: int
  mlp_dim: int
  num_heads: int
  train: bool
  dynamic_context: quant_config.DynamicContext
  dtype: Any
  dropout_rate: float
  attention_dropout_rate: float
  deterministic: bool

  # TODO(shivaniagrawal): parameterize paxis_name instead of hardcoding.
  @nn.compact
  def __call__(self,
               inputs,
               padding_mask: jnp.ndarray,
               inputs_segmentation=None):
    """Applies Encoder1DBlock module.

    Args:
      inputs: input data
      padding_mask: bool, mask padding tokens
      inputs_segmentation: input segmentation info for packed examples.

    Returns:
      output after transformer block.
    """

    # Attention block.
    batch_size, sequence_length, channel_size = inputs.shape
    shape_utils.assert_shapes_equal(padding_mask.shape,
                                    (batch_size, sequence_length, 1))
    x = aqt_flax_layers.LayerNormAqt(
        dtype=self.dtype,
        hparams=self.hparams.layer_norm,
        dynamic_context=self.dynamic_context)(
            inputs)
    x = aqt_flax_attention.SelfAttentionAqt(
        hparams=self.hparams.attention,
        num_heads=self.num_heads,
        dtype=self.dtype,
        qkv_features=self.qkv_dim,
        attention_axis=(1,),
        paxis_name='batch',
        train=self.train,
        dynamic_context=self.dynamic_context,
        causal_mask=False,
        kernel_init=nn.initializers.xavier_uniform(),
        bias_init=nn.initializers.normal(stddev=1e-6),
        use_bias=False,
        broadcast_dropout=False,
        dropout_rate=self.attention_dropout_rate,
        deterministic=self.deterministic,
        decode=False,
        name='enc_self_att')(
            x, padding_mask=padding_mask, segmentation=inputs_segmentation)

    x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=self.deterministic)
    x = x + inputs

    # MLP block.
    y = aqt_flax_layers.LayerNormAqt(
        dtype=self.dtype,
        hparams=self.hparams.layer_norm,
        dynamic_context=self.dynamic_context)(
            x)
    y = MlpBlock(
        mlp_dim=self.mlp_dim,
        hparams=self.hparams.mlp_block,
        # inputs would be signed, called after attention layer.
        train=self.train,
        dynamic_context=self.dynamic_context,
        dtype=self.dtype,
        dropout_rate=self.dropout_rate,
        deterministic=self.deterministic,
        name='mlp_block')(
            y, padding_mask=padding_mask)
    out = x + y
    shape_utils.assert_shapes_equal(out.shape,
                                    (batch_size, sequence_length, channel_size))
    return out


class EncoderDecoder1DBlock(nn.Module):
  """Transformer encoder-decoder layer.

  Attributes:
    hparams: hyperparams
    qkv_dim: dimension of the query/key/value
    mlp_dim: dimension of the mlp on top of attention block
    num_heads: number of heads
    update_bounds: Bool whether to update activation bounds.
    train: Whether model is training.
    collect_acts_stats: Whether to tag activations to record statistics.
    dtype: the dtype of the computation (default: float32)
    dropout_rate: dropout rate
    attention_dropout_rate: dropout rate for attention weights
    deterministic: bool, deterministic or not (to apply dropout)
    decode: whether to run in decoding mode
  """

  @dataclass
  class HParams:
    mlp_block: MlpBlock.HParams
    self_attention: aqt_flax_attention.MultiHeadDotProductAttentionAqt.HParams
    enc_dec_attention: aqt_flax_attention.MultiHeadDotProductAttentionAqt.HParams
    layer_norm: aqt_flax_layers.LayerNormAqt.HParams

  hparams: HParams
  qkv_dim: int
  mlp_dim: int
  num_heads: int
  train: bool
  dynamic_context: quant_config.DynamicContext
  dtype: Any
  dropout_rate: float
  attention_dropout_rate: float
  deterministic: bool
  decode: bool

  # TODO(shivaniagrawal): parametrize paxis_name instead of hardcoding.
  @nn.compact
  def __call__(
      self,
      targets,
      encoded,
      padding_mask: jnp.ndarray,
      key_padding_mask: jnp.ndarray,
      inputs_segmentation=None,
      targets_segmentation=None,
  ):
    """Applies EncoderDecoder1DBlock module.

    Args:
      targets: input data for decoder
      encoded: input data from encoder
      padding_mask: bool, mask padding tokens
      key_padding_mask: bool, mask padding tokens
      inputs_segmentation: input segmentation info for packed examples.
      targets_segmentation: target segmentation info for packed examples.

    Returns:
      output after transformer block.
    """

    # Decoder block.
    batch_size, sequence_length, num_channels = targets.shape
    encoded_sequence_length = encoded.shape[1]
    shape_utils.assert_shapes_equal(padding_mask.shape,
                                    (batch_size, sequence_length, 1))
    shape_utils.assert_shapes_equal(
        encoded.shape, (batch_size, encoded_sequence_length, num_channels))
    shape_utils.assert_shapes_equal(key_padding_mask.shape,
                                    (batch_size, encoded_sequence_length, 1))

    x = aqt_flax_layers.LayerNormAqt(
        dtype=self.dtype,
        hparams=self.hparams.layer_norm,
        dynamic_context=self.dynamic_context)(
            targets)
    x = aqt_flax_attention.SelfAttentionAqt(
        hparams=self.hparams.self_attention,
        num_heads=self.num_heads,
        dtype=self.dtype,
        qkv_features=self.qkv_dim,
        attention_axis=(1,),
        paxis_name='batch',
        train=self.train,
        dynamic_context=self.dynamic_context,
        causal_mask=True,
        kernel_init=nn.initializers.xavier_uniform(),
        bias_init=nn.initializers.normal(stddev=1e-6),
        use_bias=False,
        broadcast_dropout=False,
        dropout_rate=self.attention_dropout_rate,
        deterministic=self.deterministic,
        name='dec_self_att',
        decode=self.decode)(
            x,
            padding_mask=padding_mask,
            segmentation=targets_segmentation,
        )
    x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=self.deterministic)
    x = x + targets

    # Encoder-Decoder block.
    y = aqt_flax_layers.LayerNormAqt(
        dtype=self.dtype,
        hparams=self.hparams.layer_norm,
        dynamic_context=self.dynamic_context)(
            x)
    y = aqt_flax_attention.MultiHeadDotProductAttentionAqt(
        hparams=self.hparams.enc_dec_attention,
        num_heads=self.num_heads,
        dtype=self.dtype,
        qkv_features=self.qkv_dim,
        attention_axis=(1,),
        paxis_name='batch',
        train=self.train,
        dynamic_context=self.dynamic_context,
        causal_mask=False,
        kernel_init=nn.initializers.xavier_uniform(),
        bias_init=nn.initializers.normal(stddev=1e-6),
        use_bias=False,
        broadcast_dropout=False,
        dropout_rate=self.attention_dropout_rate,
        deterministic=self.deterministic,
        decode=False,
        name='dec_enc_att')(
            inputs_q=y,
            inputs_kv=encoded,
            padding_mask=padding_mask,
            key_padding_mask=key_padding_mask,
            segmentation=targets_segmentation,
            key_segmentation=inputs_segmentation)
    y = nn.Dropout(rate=self.dropout_rate)(y, deterministic=self.deterministic)
    y = y + x

    # MLP block.
    z = aqt_flax_layers.LayerNormAqt(
        dtype=self.dtype,
        hparams=self.hparams.layer_norm,
        dynamic_context=self.dynamic_context)(
            y)
    z = MlpBlock(
        mlp_dim=self.mlp_dim,
        dtype=self.dtype,
        hparams=self.hparams.mlp_block,
        # inputs would be signed, called after attention layer.
        train=self.train,
        dynamic_context=self.dynamic_context,
        dropout_rate=self.dropout_rate,
        deterministic=self.deterministic,
        name='mlp_block')(
            z, padding_mask=padding_mask)

    return y + z


class Encoder(nn.Module):
  """Transformer Model Encoder for sequence to sequence translation.

  Attributes:
    vocab_size: size of the vocabulary
    inputs_segmentation: input segmentation info for packed examples.
    targets_segmentation: target segmentation info for packed examples.
    shared_embedding: a shared embedding layer to use.
    use_bfloat16: bool: whether use bfloat16.
    emb_dim: dimension of embedding
    num_heads: number of heads
    qkv_dim: dimension of the query/key/value
    mlp_dim: dimension of the mlp on top of attention block
    train: if it is training,
    collect_acts_stats: Whether to tag activations to record statistics.
    dropout_rate: dropout rate
    attention_dropout_rate: dropout rate for attention weights
  """

  @dataclass
  class HParams:
    """Encoder hyperparameters."""

    embedding: aqt_flax_layers.EmbedAqt.HParams
    encoder_1d_blocks: Tuple[Encoder1DBlock.HParams, ...]
    layer_norm: aqt_flax_layers.LayerNormAqt.HParams

    # TODO(b/175245107): Remove this function
    @classmethod
    def create_from_block_template(
        cls: Type[T], *, embedding: aqt_flax_layers.EmbedAqt.HParams,
        block_template: Encoder1DBlock.HParams,
        layer_norm: aqt_flax_layers.LayerNormAqt.HParams, num_layers: int) -> T:
      return cls(
          embedding=embedding,
          layer_norm=layer_norm,
          encoder_1d_blocks=tuple(
              copy.deepcopy(block_template) for _ in range(num_layers)))

  vocab_size: int
  hparams: HParams
  max_len: int
  shared_embedding: Any
  use_bfloat16: bool
  emb_dim: int
  num_heads: int
  qkv_dim: int
  mlp_dim: int
  train: bool
  dynamic_context: quant_config.DynamicContext
  dropout_rate: float
  attention_dropout_rate: float

  # TODO(shivaniagrawal): parametrize paxis_name instead of hardcoding.
  @nn.compact
  def __call__(self, inputs, inputs_positions=None, inputs_segmentation=None):
    """Applies Transformer model on the inputs.

    Args:
      inputs: input data
      inputs_positions: input subsequence positions for packed examples.
      inputs_segmentation: input segmentation info for packed examples.

    Returns:
      output of a transformer decoder.

    """
    batch_size, sequence_length = inputs.shape

    # Padding Masks
    src_padding_mask = (inputs > 0)[..., None]
    shape_utils.assert_shapes_equal(src_padding_mask.shape,
                                    (batch_size, sequence_length, 1))

    if self.use_bfloat16:
      dtype = jnp.bfloat16
    else:
      dtype = jnp.float32

    # Input Embedding
    if self.shared_embedding is None:
      input_embed = aqt_flax_layers.EmbedAqt(
          num_embeddings=self.vocab_size,
          features=self.emb_dim,
          hparams=self.hparams.embedding,
          embedding_init=nn.initializers.normal(stddev=self.emb_dim**-0.5),
          dtype=dtype,
          name='input_embed',
          paxis_name='batch',
          train=self.train,
          dynamic_context=self.dynamic_context)
    else:
      input_embed = self.shared_embedding
    x = inputs.astype('int32')
    x = input_embed(x) * jnp.sqrt(self.emb_dim)
    x = AddPositionEmbs(
        name='posembed_input',
        max_len=self.max_len,
        min_timescale=1.0,
        max_timescale=10000.0,
        decode=False)(
            x, inputs_positions=inputs_positions)
    x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=not self.train)

    if self.use_bfloat16:
      x = x.astype(jnp.bfloat16)

    # Input Encoder
    num_layers = len(self.hparams.encoder_1d_blocks)
    for lyr in range(num_layers):
      x = Encoder1DBlock(
          train=self.train,
          dynamic_context=self.dynamic_context,
          qkv_dim=self.qkv_dim,
          mlp_dim=self.mlp_dim,
          num_heads=self.num_heads,
          hparams=self.hparams.encoder_1d_blocks[lyr],
          dtype=dtype,
          dropout_rate=self.dropout_rate,
          attention_dropout_rate=self.attention_dropout_rate,
          deterministic=not self.train,
          name=f'encoderblock_{lyr}')(
              x,
              padding_mask=src_padding_mask,
              inputs_segmentation=inputs_segmentation)
    encoded = aqt_flax_layers.LayerNormAqt(
        dtype=dtype,
        name='encoder_norm',
        hparams=self.hparams.layer_norm,
        dynamic_context=self.dynamic_context)(
            x)
    shape_utils.assert_shapes_equal(encoded.shape,
                                    (batch_size, sequence_length, self.emb_dim))
    return encoded


class Decoder(nn.Module):
  """Transformer Model Decoder for sequence to sequence translation.

  Attributes:
    output_vocab_size: size of the vocabulary.
    update_bounds: Bool whether to update activation bounds.
    hparams: hyperparameters
    max_len: maximum length.
    shared_embedding: a shared embedding layer to use.
    logits_via_embedding: bool: whether final logit transform shares embedding
      weights.
    use_bfloat16: bool: whether use bfloat16.
    emb_dim: dimension of embedding
    num_heads: number of heads
    qkv_dim: dimension of the query/key/value
    mlp_dim: dimension of the mlp on top of attention block
    train: if it is training.
    collect_acts_stats: Whether to tag activations to record statistics.
    dropout_rate: dropout rate
    attention_dropout_rate: dropout rate for attention weights
    paxis_name: name of parallel reduction axis
    decode: whether to run in decoding mode.
  """

  @dataclass
  class HParams:
    """Decoder hyperparameters."""

    embedding: aqt_flax_layers.EmbedAqt.HParams
    encoder_decoder_1d_blocks: Tuple[EncoderDecoder1DBlock.HParams, ...]
    # logits field must be set if logits are not shared.
    logits: Optional[aqt_flax_layers.DenseAqt.HParams]
    layer_norm: aqt_flax_layers.LayerNormAqt.HParams

    # TODO(b/175245107): Remove this function
    @classmethod
    def create_from_block_template(
        cls: Type[T],
        *,
        embedding: aqt_flax_layers.EmbedAqt.HParams,
        block_template: EncoderDecoder1DBlock.HParams,
        num_layers: int,
        layer_norm: aqt_flax_layers.LayerNormAqt.HParams,
        logits: Optional[aqt_flax_layers.DenseAqt.HParams] = None) -> T:
      return cls(
          embedding=embedding,
          layer_norm=layer_norm,
          encoder_decoder_1d_blocks=tuple(
              copy.deepcopy(block_template) for _ in range(num_layers)),
          logits=logits)

  output_vocab_size: int
  hparams: HParams
  max_len: int
  shared_embedding: Any
  logits_via_embedding: bool
  use_bfloat16: bool
  emb_dim: int
  num_heads: int
  qkv_dim: int
  mlp_dim: int
  train: bool
  dynamic_context: quant_config.DynamicContext
  dropout_rate: float
  attention_dropout_rate: float
  paxis_name: Optional[str]
  decode: bool

  # TODO(shivaniagrawal): parametrize paxis_name instead of hardcoding.
  @nn.compact
  def __call__(
      self,
      encoded,
      src_padding_mask,
      targets,
      targets_positions=None,
      inputs_segmentation=None,
      targets_segmentation=None,
      tgt_padding_mask=None,
  ):
    """Applies Transformer model on the inputs.

    Args:
      encoded: encoded input data from encoder.
      src_padding_mask: padding mask for inputs.
      targets: target inputs.
      targets_positions: input subsequence positions for packed examples.
      inputs_segmentation: input segmentation info for packed examples.
      targets_segmentation: target segmentation info for packed examples.
      tgt_padding_mask: target tokens padding mask.

    Returns:
      output of a transformer decoder.

    """
    batch_size, sequence_length, channel_size = encoded.shape  # pylint: disable=unused-variable
    target_batch_size, target_sequence_length = targets.shape  # pylint: disable=unused-variable
    shape_utils.assert_shapes_equal(targets.shape,
                                    (batch_size, target_sequence_length))

    # Padding Masks
    if tgt_padding_mask is None:
      tgt_padding_mask = (targets > 0)[..., None]
    shape_utils.assert_shapes_equal(tgt_padding_mask.shape,
                                    (batch_size, target_sequence_length, 1))

    if self.use_bfloat16:
      dtype = jnp.bfloat16
    else:
      dtype = jnp.float32

    # Target Embedding
    if self.shared_embedding is None:
      output_embed = aqt_flax_layers.EmbedAqt(
          num_embeddings=self.output_vocab_size,
          features=self.emb_dim,
          hparams=self.hparams.embedding,
          embedding_init=nn.initializers.normal(stddev=self.emb_dim**-0.5),
          dtype=dtype,
          name='target_embed',
          train=self.train,
          dynamic_context=self.dynamic_context,
          paxis_name='batch')
    else:
      output_embed = self.shared_embedding

    y = targets.astype('int32')
    if not self.decode:
      y = shift_right(y)
    y = output_embed(y) * jnp.sqrt(self.emb_dim)
    y = AddPositionEmbs(
        name='posembed_targets',
        max_len=self.max_len,
        decode=self.decode,
        min_timescale=1.0,
        max_timescale=10000.0)(
            y, inputs_positions=targets_positions)
    y = nn.Dropout(rate=self.dropout_rate)(y, deterministic=not self.train)

    if self.use_bfloat16:
      y = y.astype(jnp.bfloat16)

    # Target-Input Decoder
    num_layers = len(self.hparams.encoder_decoder_1d_blocks)
    for lyr in range(num_layers):
      y = EncoderDecoder1DBlock(
          train=self.train,
          dynamic_context=self.dynamic_context,
          qkv_dim=self.qkv_dim,
          mlp_dim=self.mlp_dim,
          num_heads=self.num_heads,
          hparams=self.hparams.encoder_decoder_1d_blocks[lyr],
          dtype=dtype,
          dropout_rate=self.dropout_rate,
          attention_dropout_rate=self.attention_dropout_rate,
          deterministic=not self.train,
          name=f'encoderdecoderblock_{lyr}',
          decode=self.decode)(
              y,
              encoded,
              padding_mask=tgt_padding_mask,
              key_padding_mask=src_padding_mask,
              inputs_segmentation=inputs_segmentation,
              targets_segmentation=targets_segmentation)
    y = aqt_flax_layers.LayerNormAqt(
        dtype=dtype,
        name='encoderdecoder_norm',
        hparams=self.hparams.layer_norm,
        dynamic_context=self.dynamic_context)(
            y)
    y = y.reshape((batch_size * target_sequence_length, channel_size))
    tgt_padding_mask = tgt_padding_mask.reshape(
        (batch_size * target_sequence_length, 1))
    # Decoded Logits
    if self.logits_via_embedding:
      # Use the transpose of embedding matrix for logit transform.
      logits = output_embed.attend(
          query=y,
          padding_mask=tgt_padding_mask,
          paxis_name=self.paxis_name,
          train=self.train)
    else:
      if self.hparams.logits is None:
        raise ValueError('If logits_via_embedding is False, then the hparams '
                         'for the logits layer have to be provided.')
      logits = aqt_flax_layers.DenseAqt(
          features=self.output_vocab_size,
          dtype=dtype,
          paxis_name='batch',
          train=self.train,
          dynamic_context=self.dynamic_context,
          hparams=self.hparams.logits,
          kernel_init=nn.initializers.xavier_uniform(),
          bias_init=nn.initializers.normal(stddev=1e-6),
          name='logits_dense')(
              y, padding_mask=tgt_padding_mask)
    return logits


# The following final model is simple but looks verbose due to all the
# repetitive keyword argument plumbing.  It just sticks the Encoder and
# Decoder in series for training, but allows running them separately for
# inference.
# TODO(malmaud): Remove this comment once the new hyperparam threading
# logic is complete.


class Transformer(nn.Module):
  """Transformer Model for sequence to sequence translation.

  Attributes:
    update_bounds: Bool whether to update activation bounds.
    hparams: hyperparameters
    max_len: maximum length.
    vocab_size: size of the input vocabulary
    output_vocab_size: size of the output vocabulary
    use_bfloat16: bool: whether use bfloat16.
    train: if it is training,
    collect_acts_stats: Whether to tag activations to record statistics.
    dropout_rate: dropout rate
    attention_dropout_rate: dropout rate for attention weights
    should_decode: whether to run in decoding mode
  """

  @dataclass
  class HParams:
    """Hyperparameters for the transformer.

    TODO(wanglisa): Push encoder / decoder specific params (num_heads, emb_dim
    qkv_dim, mlp_dim) down to Encoder.HParams and Decoder.HParams.
    """

    encoder: Encoder.HParams
    decoder: Decoder.HParams
    emb_dim: int  # dimension of embedding.
    num_heads: int  # Number of heads.
    qkv_dim: int  # dimension of the query/key/value.
    mlp_dim: int  # dimension of the mlp on top of attention block.
    share_embeddings: bool  # share embedding layer for inputs and targets.
    logits_via_embedding: bool  # share emb weights w/ final logit transform.

  hparams: HParams
  max_len: int
  vocab_size: Optional[int]
  output_vocab_size: Optional[int]
  use_bfloat16: bool
  train: bool
  dynamic_context: quant_config.DynamicContext
  dropout_rate: float
  attention_dropout_rate: float
  # We call this 'should_decode' instead of 'decode' so it doesn't conflict with
  # the 'decode' method name.
  should_decode: bool

  def setup(self):

    if self.use_bfloat16:
      dtype = jnp.bfloat16
    else:
      dtype = jnp.float32

    if self.hparams.share_embeddings:
      if self.output_vocab_size is not None:
        assert self.output_vocab_size == self.vocab_size, (
            "can't share embedding with different vocab sizes.")
      self.shared_embedding = aqt_flax_layers.EmbedAqt(  # pylint: disable=missing-from-attributes
          num_embeddings=self.vocab_size,
          features=self.hparams.emb_dim,
          hparams=self.hparams.encoder.embedding,
          dtype=dtype,
          embedding_init=nn.initializers.normal(
              stddev=self.hparams.emb_dim**-0.5),
          train=self.train,
          dynamic_context=self.dynamic_context,
          paxis_name='batch')
    else:
      self.shared_embedding = None

    self.encoder = Encoder(  # pylint: disable=missing-from-attributes
        hparams=self.hparams.encoder,
        vocab_size=self.vocab_size,
        shared_embedding=self.shared_embedding,
        use_bfloat16=self.use_bfloat16,
        emb_dim=self.hparams.emb_dim,
        num_heads=self.hparams.num_heads,
        qkv_dim=self.hparams.qkv_dim,
        mlp_dim=self.hparams.mlp_dim,
        max_len=self.max_len,
        train=self.train,
        dynamic_context=self.dynamic_context,
        dropout_rate=self.dropout_rate,
        attention_dropout_rate=self.attention_dropout_rate,
    )

    self.decoder = Decoder(  # pylint: disable=missing-from-attributes
        hparams=self.hparams.decoder,
        output_vocab_size=self.output_vocab_size,
        shared_embedding=self.shared_embedding,
        logits_via_embedding=self.hparams.logits_via_embedding,
        use_bfloat16=self.use_bfloat16,
        emb_dim=self.hparams.emb_dim,
        num_heads=self.hparams.num_heads,
        qkv_dim=self.hparams.qkv_dim,
        mlp_dim=self.hparams.mlp_dim,
        max_len=self.max_len,
        train=self.train,
        dynamic_context=self.dynamic_context,
        dropout_rate=self.dropout_rate,
        attention_dropout_rate=self.attention_dropout_rate,
        paxis_name='batch',
        decode=self.should_decode)

  # TODO(shivaniagrawal): parametrize paxis_name instead of hardcoding.
  def __call__(
      self,
      inputs,
      targets,
      inputs_positions=None,
      targets_positions=None,
      inputs_segmentation=None,
      targets_segmentation=None,
  ):
    """Applies Transformer model on the inputs.

    Args:
      inputs: input data
      targets: target data
      inputs_positions: input subsequence positions for packed examples.
      targets_positions: target subsequence positions for packed examples.
      inputs_segmentation: input segmentation info for packed examples.
      targets_segmentation: target segmentation info for packed examples.

    Returns:
      output of a transformer decoder.

    """
    batch_size, sequence_length = inputs.shape
    assert batch_size == targets.shape[
        0], 'Input and targets must have the same batch size'

    src_padding_mask = (inputs > 0)[..., None]
    shape_utils.assert_shapes_equal(src_padding_mask.shape,
                                    (batch_size, sequence_length, 1))

    encoded = self.encode(
        inputs,
        inputs_positions=inputs_positions,
        inputs_segmentation=inputs_segmentation)

    logits = self.decode(
        encoded,
        src_padding_mask,
        targets,
        targets_positions=targets_positions,
        inputs_segmentation=inputs_segmentation,
        targets_segmentation=targets_segmentation,
        tgt_padding_mask=None)
    return logits.astype(jnp.float32) if self.use_bfloat16 else logits

  # The following two methods allow us to run the trained Transformer in
  # two parts during fast decoding.  First, we call the encoder branch to
  # encode the inputs, then we call the decoder branch while providing a
  # cache object for iteratively storing keys and values during the decoding
  # process.

  # TODO(shivaniagrawal): parametrize paxis_name instead of hardcoding.
  def encode(self, inputs, inputs_positions=None, inputs_segmentation=None):

    encoded = self.encoder(
        inputs,
        inputs_positions=inputs_positions,
        inputs_segmentation=inputs_segmentation)

    return encoded

  # TODO(shivaniagrawal): parametrize paxis_name instead of hardcoding.
  def decode(
      self,
      encoded,
      src_padding_mask,
      targets,
      targets_positions=None,
      inputs_segmentation=None,
      targets_segmentation=None,
      tgt_padding_mask=None,
  ):
    """Run the decoder part of the transformer given encoded inputs."""
    logits = self.decoder(
        encoded,
        src_padding_mask,
        targets,
        targets_positions=targets_positions,
        inputs_segmentation=inputs_segmentation,
        targets_segmentation=targets_segmentation,
        tgt_padding_mask=tgt_padding_mask,
    )

    return logits



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for config_schema.py for Imagenet."""

import json

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax.imagenet.configs_script import config_schema


class BaseConfigTest(parameterized.TestCase):

  def test_precision_propagates(self):
    config = config_schema.get_config(num_blocks=16, use_auto_acts=True)

    # Set the global precision to 4 bits.
    config.prec = 4
    # Set the global half_shift flag to False
    config.half_shift = False
    # Test that this sets the weight and activation to 4 as well.
    self.assertEqual(config.weight_prec, 4)
    self.assertEqual(config.quant_act.prec, 4)
    # Test that this sets the weight_half_shift and act half_shift to False
    self.assertEqual(config.weight_half_shift, False)
    self.assertEqual(config.quant_act.half_shift, False)
    # Test that propagates all the way down to the weight precision of layer
    # types and individual layers. As an example of an individual layer, we take
    # the dense1 matmul of the second block of the decoder.
    conv1_block3 = config.model_hparams.residual_blocks[2].conv_1
    # Meanwhile, 'conv1' represents the generic configuration of all conv1
    # layers throughout the model.
    conv1 = config.residual.conv_1
    self.assertEqual(conv1.weight_prec, 4)
    self.assertEqual(conv1_block3.weight_prec, 4)
    self.assertEqual(conv1.weight_half_shift, False)
    self.assertEqual(conv1.quant_act.half_shift, False)
    self.assertEqual(conv1_block3.weight_half_shift, False)
    self.assertEqual(conv1_block3.quant_act.half_shift, False)

    # Test if we take the same config instance and alter the global precision to
    # 8, it automatically propagates to individual layers.
    config.prec = 8
    self.assertEqual(conv1.weight_prec, 8)
    self.assertEqual(conv1_block3.weight_prec, 8)
    # Test if we take the same config instance and alter the global half_shift
    # to True, it automatically propagates to individual layers.
    config.half_shift = True
    self.assertEqual(conv1.weight_half_shift, True)
    self.assertEqual(conv1.quant_act.half_shift, True)
    self.assertEqual(conv1_block3.weight_half_shift, True)
    self.assertEqual(conv1_block3.quant_act.half_shift, True)

    # Test that the precision can be overridden for a specific layer type. We
    # want to verify that the change doesn't back-propagate back to the global
    # precision field but does propagate down to individual layers of that layer
    # type. We only want changes to fields to automatically propagate down the
    # parameter hierarchy, not up.
    conv1.weight_prec = 2
    self.assertEqual(conv1.weight_prec, 2)
    self.assertEqual(conv1_block3.weight_prec, 2)
    self.assertEqual(config.prec, 8)

    # Now update the precision for just a specific layer and check that it
    # doesn't propagate upwards.
    conv1_block3.weight_prec = 1
    self.assertEqual(conv1_block3.weight_prec, 1)
    self.assertEqual(conv1.weight_prec, 2)
    self.assertEqual(config.prec, 8)

  @parameterized.parameters(dict(num_blocks=10), dict(num_blocks=20))
  def test_num_blocks(self, num_blocks):
    config = config_schema.get_config(num_blocks=num_blocks, use_auto_acts=True)
    self.assertLen(config.model_hparams.residual_blocks, num_blocks)
    self.assertLen(config.model_hparams.residual_blocks, num_blocks)

  def test_auto_acts_parameter(self):
    # If use_auto_acts is False, then the bounds should be a single scalar that
    # specifies the fixed bound; 'None' by default.
    config = config_schema.get_config(num_blocks=15, use_auto_acts=False)
    self.assertIsNone(config.quant_act.bounds)
    # If use_auto_acts is True, it should have the same structure as the
    # GetBounds.Hyper dataclass.
    config = config_schema.get_config(num_blocks=15, use_auto_acts=True)
    self.assertIn('initial_bound', config.quant_act.bounds)

    # Because the config dict is locked, it shouldn't be possible to change it
    # back to fixed bounds if it was created with use_auto_acts=True.
    with self.assertRaises(TypeError):
      config.quant_act.bounds = 1.0

  @parameterized.parameters(dict(num_blocks=16), dict(num_blocks=33))
  def test_schema_matches_expected(self, num_blocks):
    # This tests that the schema of the configdict returned by 'config_schema',
    # once all references are resolved, matches an expected schema. 'Schema'
    # here means the names and structure of fields at each level of the
    # configuration hierarchy. A value of 'None' in the expected schemas defined
    # below indicates a real configuration would have a concrete scalar value
    # there.
    quant_act_schema = {
        'bounds': {
            'initial_bound': None,
            'stddev_coeff': None,
            'absdev_coeff': None,
            'mix_coeff': None,
            'reset_stats': None,
            'ema_coeff': None,
            'use_cams': None,
            'exclude_zeros': None,
            'use_mean_of_max': None,
            'granularity': None,
            'fixed_bound': None,
            'cams_coeff': None,
            'cams_stddev_coeff': None,
            'mean_of_max_coeff': None,
            'use_old_code': None,
        },
        'input_distribution': None,
        'prec': None,
        'half_shift': None,
    }

    dense_schema = {
        'weight_prec': None,
        'weight_quant_granularity': None,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'weight_half_shift': None,
        'weight_sparsity': sparsity_schema,
        'act_sparsity': sparsity_schema,
    }

    conv_schema = {
        'weight_prec': None,
        'weight_quant_granularity': None,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'weight_half_shift': None,
    }

    residual_block_schema = {
        'conv_se': conv_schema,
        'conv_proj': conv_schema,
        'conv_1': conv_schema,
        'conv_2': conv_schema,
        'conv_3': conv_schema,
        'act_function': None,
        'shortcut_ch_expand_method': None,
        'shortcut_ch_shrink_method': None,
        'shortcut_spatial_method': None,
    }

    expected_top_level_schema = {
        'metadata': {
            'description': None,
            'hyper_str': None
        },
        'base_learning_rate': None,
        'momentum': None,
        'weight_decay': None,
        'activation_bound_update_freq': None,
        'activation_bound_start_step': None,
        'weight_quant_start_step': None,
        'prec': None,
        'half_shift': None,
        'weight_prec': None,
        'weight_half_shift': None,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'weight_quant_granularity': None,
        'early_stop_steps': None,
        'act_function': None,
        'shortcut_ch_shrink_method': None,
        'shortcut_ch_expand_method': None,
        'shortcut_spatial_method': None,
        'teacher_model': None,
        'is_teacher': None,
        'seed': None,
        'sparsity': sparsity_schema,
        'weight_sparsity': sparsity_schema,
        'act_sparsity': sparsity_schema,
        'lr_scheduler': {
            'warmup_epochs': None,
            'cooldown_epochs': None,
            'scheduler': None,
            'num_epochs': None,
            'endlr': None,
            'knee_lr': None,
            'knee_epochs': None,
        },
        'optimizer': None,
        'adam': {
            'beta1': None,
            'beta2': None,
        },
        'dense_layer': dense_schema,
        'conv': conv_schema,
        'residual': residual_block_schema,
        'model_hparams': {
            'dense_layer': dense_schema,
            'conv_init': conv_schema,
            'residual_blocks': [residual_block_schema] * num_blocks,
            'filter_multiplier': None,
            'act_function': None,
            'se_ratio': None,
            'init_group': None,
        },
    }

    config = config_schema.get_config(num_blocks=num_blocks, use_auto_acts=True)
    # This round-trip conversion from JSON forces all references to resolve to
    # concrete values.
    config_reified = json.loads(config.to_json())

    # This test is not interested in checking the specific values of fields in
    # the configuration, but only that the schema of the hierarchies
    # are the same. Thus we all set the value of leaf nodes in the config to
    # 'None' before checking that the actual and expected configuration
    # structures are the same.
    def set_leaves_to_none(config):
      # We are at an intermediate node in the tree-structured input, which could
      # either be in the form of a dictionary or a list of other nodes in the
      # tree.
      if isinstance(config, dict):
        return {key: set_leaves_to_none(value) for key, value in config.items()}
      elif isinstance(config, list):
        return [set_leaves_to_none(value) for value in config]

      # We are at a leaf node in the tree-structured input.
      else:
        return None

    self.assertSameStructure(
        set_leaves_to_none(config_reified), expected_top_level_schema)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates ConfigDict instances for the WMT transformer.

Most users will just create a ConfigDict instance with 'get_config' and then
override its parameters to specialize the configuration.
"""

import copy
import enum
from typing import Optional
from aqt.jax_legacy.utils import config_schema_utils
import ml_collections


float_ph = config_schema_utils.float_ph
int_ph = config_schema_utils.int_ph
str_ph = config_schema_utils.str_ph
bool_ph = config_schema_utils.bool_ph


def get_wmt_base_config(use_auto_acts: bool,
                        fp_quant: bool) -> ml_collections.ConfigDict:
  """Return a base ConfigDict which does not yet have fields for individual layers."""
  base_config = config_schema_utils.get_base_config(
      use_auto_acts, fp_quant=fp_quant)
  base_config.update({
      "learning_rate_schedule": {
          "factors": str_ph(),
          "base_learning_rate": float_ph(),
          "warmup_steps": int_ph(),
          "decay_factor": float_ph(),
          "steps_per_decay": int_ph(),
          "steps_per_cycle": int_ph(),
      },
      "per_host_batch_size": int_ph(),
      "num_train_steps": int_ph(),
      "beta1": float_ph(),
      "beta2": float_ph(),
      "eps": float_ph(),
      "random_seed": int_ph(),
      "hardware_rng": bool_ph(),
      "weight_outlier_regularization": float_ph(),
      "weight_outlier_regularization_regex": str_ph(),
      "prefer_int8_to_int32_dot": bool_ph(),
      "model_hparams": {
          "emb_dim": int_ph(),
          "num_heads": int_ph(),
          "qkv_dim": int_ph(),
          "mlp_dim": int_ph(),
          "share_embeddings": bool_ph(),
          "logits_via_embedding": bool_ph(),
      },
  })

  base_config.dense = config_schema_utils.get_dense_config(base_config)
  base_config.mlp_block = get_mlp_block_config(base_config)
  base_config.embedding = get_embedding_config(base_config)
  base_config.attention = get_attention_config(base_config)

  return base_config


def get_mlp_block_config(
    parent_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Create a ConfigDict corresponding to wmt_mlperf.models.MlpBlock.HParams."""
  config = ml_collections.ConfigDict()
  config_schema_utils.set_default_reference(
      config, parent_config, ["dense_1", "dense_2"], parent_field="dense")
  config.dense_2.quant_act.input_distribution = "positive"
  config.lock()
  return config


def get_fp_quant_hparams_config(
    quantized: bool,
    quantized_reductions: bool) -> Optional[ml_collections.ConfigDict]:
  """Create appropriate setting for 'QuantHParams' field used by softmax and LayerNormAQT."""
  if quantized_reductions and not quantized:
    raise ValueError("If `quantized` is False, `quantized_reductions` "
                     "must also be False.")
  if quantized:
    quant_hparams = ml_collections.ConfigDict({
        "prec":
            config_schema_utils.get_fp_config(),
        "reduction_prec":
            config_schema_utils.get_fp_config()
            if quantized_reductions else None
    })
  else:
    quant_hparams = None
  return quant_hparams


def get_softmax_config(quantized: bool,
                       quantized_reductions: bool) -> ml_collections.ConfigDict:
  """Create a ConfigDIct corresponding to flax_attention.SoftmaxHParams."""
  quant_hparams = get_fp_quant_hparams_config(
      quantized=quantized, quantized_reductions=quantized_reductions)
  config = ml_collections.ConfigDict({
      "exp_hparams": None,
      "reciprocal_hparams": None,
      "quant_hparams": quant_hparams
  })
  config.lock()
  return config


def get_layer_norm_config(
    quantized: bool, quantized_reductions: bool) -> ml_collections.ConfigDict:
  """Create a ConfigDict corresponding to flax_layers.LayerNormAqt.HParams."""
  quant_hparams = get_fp_quant_hparams_config(
      quantized=quantized, quantized_reductions=quantized_reductions)
  config = ml_collections.ConfigDict({"quant_hparams": quant_hparams})
  config.lock()
  return config


def set_global_softmax_config(
    base_config: ml_collections.ConfigDict,
    softmax_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Sets all layers' softmax configuration in `base_config` to `softmax_config`."""

  def set_softmax(cfg):
    # We unlock the configs since the 'softmax' field might not already exist.
    with cfg.unlocked():
      cfg.softmax = copy.deepcopy(softmax_config)

  base_config = copy.deepcopy(base_config)
  for block in base_config.model_hparams.encoder.encoder_1d_blocks:
    set_softmax(block.attention.attn_acts)
  for block in base_config.model_hparams.decoder.encoder_decoder_1d_blocks:
    set_softmax(block.self_attention.attn_acts)
    set_softmax(block.enc_dec_attention.attn_acts)
  return base_config


def set_global_layer_norm_config(
    base_config: ml_collections.ConfigDict,
    layer_norm_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Sets all layers' softmax configuration in `base_config` to `layer_norm_config`."""

  def set_layer_norm(cfg):
    # We unlock the configs since the 'layer_norm' field might not already
    # exist.
    with cfg.unlocked():
      cfg.layer_norm = copy.deepcopy(layer_norm_config)

  base_config = copy.deepcopy(base_config)
  set_layer_norm(base_config.model_hparams.encoder)
  set_layer_norm(base_config.model_hparams.decoder)
  for block in base_config.model_hparams.encoder.encoder_1d_blocks:
    set_layer_norm(block)
  for block in base_config.model_hparams.decoder.encoder_decoder_1d_blocks:
    set_layer_norm(block)
  return base_config


def get_attention_config(
    parent_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Create a ConfigDict corresponding to aqt.flax_attention.MutliHeadDotProductAttentionAqt.HParams."""
  config = ml_collections.ConfigDict()
  config_schema_utils.set_default_reference(
      config, parent_config, ["dense_kqv", "dense_out"], parent_field="dense")

  config.attn_acts = ml_collections.ConfigDict({})

  config_schema_utils.set_default_reference(config, parent_config,
                                            ["quant_type", "quant_act"])
  config_schema_utils.set_default_reference(config.attn_acts, config,
                                            ["quant_type"])
  config_schema_utils.set_default_reference(
      config.attn_acts,
      config, ["attn_act_q", "attn_act_k", "attn_act_v"],
      parent_field="quant_act")
  config.attn_acts.attn_act_probs = ml_collections.ConfigDict({
      "input_distribution": "positive",
      "bounds": 1.0,
      "half_shift": False,  # Set half_shift to false for positive distribution
  })
  config_schema_utils.set_default_reference(config.attn_acts.attn_act_probs,
                                            parent_config.quant_act, "prec")
  config.lock()
  return config


class BlockKind(enum.Enum):
  """Enum to distinguish between an encoder and a decoder block."""
  ENCODER = enum.auto()
  DECODER = enum.auto()


def get_block_config(parent_config: ml_collections.ConfigDict,
                     block_kind: BlockKind) -> ml_collections.ConfigDict:
  """Create a ConfigDict corresponding to wmt_mlperf.models.Encoder[Decoder]1DBlock.HParams."""
  config = ml_collections.ConfigDict()
  config_schema_utils.set_default_reference(config, parent_config, "mlp_block")
  if block_kind == BlockKind.ENCODER:
    config_schema_utils.set_default_reference(config, parent_config,
                                              "attention")
  elif block_kind == BlockKind.DECODER:
    config_schema_utils.set_default_reference(
        config,
        parent_config, ["self_attention", "enc_dec_attention"],
        parent_field="attention")
  else:
    raise ValueError(f"Unknown block_kind {block_kind}")
  config.lock()
  return config


def get_embedding_config(
    parent_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Create a ConfigDict corresponding to aqt.flax_layers.Embedding.HParams."""
  config = ml_collections.ConfigDict()
  config_schema_utils.set_default_reference(
      config, parent_config,
      ["weight_prec", "quant_type", "quant_act", "weight_half_shift",
       ])
  config.lock()
  return config


def get_config(n_layers: int, use_auto_acts: bool,
               fp_quant: bool) -> ml_collections.ConfigDict:
  """Returns a ConfigDict instance for a WMT transformer.

  The ConfigDict is wired up so that changing a field at one level of the
  hierarchy changes the value of that field everywhere downstream in the
  hierarchy. For example, changing the top-level 'prec' parameter
  (eg, config.prec=4) will cause the precision of all layers to change.
  Changing the precision of a specific layer type
  (eg, config.mlp_block.dense_1.weight_prec=4) will cause the weight precision
  of all Dense1 layers to change, overriding the value of the global
  config.prec value.

  See config_schema_test.test_schema_matches_expected to see the structure
  of the ConfigDict instance this will return.

  Args:
    n_layers: Number of layers in the encoder and the decoder.
    use_auto_acts: Whether to use automatic clipping bounds for activations or
      fixed bounds. Unlike other properties of the configuration which can be
      overridden directly in the ConfigDict instance, this affects the immutable
      schema of the ConfigDict and so has to be specified before the ConfigDict
      is created.
    fp_quant: Whether to use floating point quantization. Defaults to False for
      integer quantization.

  Returns:
    A ConfigDict instance which parallels the hierarchy of TrainingHParams.
  """
  base_config = get_wmt_base_config(
      use_auto_acts=use_auto_acts, fp_quant=fp_quant)
  model_hparams = base_config.model_hparams
  model_hparams.encoder = {
      "encoder_1d_blocks": [
          get_block_config(base_config, BlockKind.ENCODER)
          for _ in range(n_layers)
      ]
  }
  config_schema_utils.set_default_reference(model_hparams.encoder, base_config,
                                            "embedding")
  model_hparams.decoder = {
      "encoder_decoder_1d_blocks": [
          get_block_config(base_config, BlockKind.DECODER)
          for _ in range(n_layers)
      ]
  }
  config_schema_utils.set_default_reference(model_hparams.decoder, base_config,
                                            "embedding")

  config_schema_utils.set_default_reference(
      model_hparams.decoder, base_config, "logits", parent_field="dense")
  base_config.lock()
  return base_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantization utils for tensors."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test
import typing
from typing import Any, Callable, Optional, Sequence, TypeAlias
from aqt.jax.v2 import utils
import flax.cursor
import flax.struct
import jax
import jax.numpy as jnp
import jax.typing as jax_typing
from typing_extensions import Self  # for python version < 3.11

GradientFn = Callable[..., Any] | None  # None when there is no numerics
_MSG_NO_QVALUE = (
    'QTensor does not have qvalue, but it is asked to access the qvalue.'
    ' Call QTensor.quant() before using the qvalue.'
)

if typing.TYPE_CHECKING:
  # These are needed to avoid static typing complaints for sharding.
  # Concatenating sharding types, e. g. jnp.ndarray | jax.sharding.NamedSharding
  # | jax.sharding.PartitionSpec, will trigger an error to logics where the
  # values are used as jax.array, since they could be sharding.
  ArrayT: TypeAlias = Any
else:
  ArrayT: TypeAlias = jnp.ndarray


@utils.flax_slots_dataclass
class QTensor:
  """Quantized tensor."""

  # Quantized (compressed) representation of tensor.
  # Use dequant() method to "decompress" to the original tensor.
  qvalue: Optional[ArrayT]

  # (scale == None) means that scale is unknown/invalid;
  # Otherwise, check dequant(self) for semantics.
  scale: Optional[list[ArrayT]]

  # Used in dot_general, transposed scales used in post dot_general scaling.
  # The same comments apply as to scale.
  # We currently keep it here because:
  # - we store scale_t in the checkpoint to avoid transposition per inference.
  # - scale_t is used both in backprop of dot_general and in post-scaling.
  #   We avoid transposing scale twice.
  # TODO(lew): Move scale_t from QTensor to some dot-general specific type?
  scale_t: Optional[list[ArrayT]]

  # DType of the tensor before quantized.
  dequant_dtype: Optional[jnp.dtype] = flax.struct.field(
      pytree_node=False, default=None
  )

  def quant(self, x):
    assert self.qvalue is None, 'Already quantized QTensor.'
    assert self.scale is not None, 'Missing scales to be used for quantization.'

    qvalue = x
    for s in self.scale:
      qvalue = qvalue * jax.lax.reciprocal(s)

    # TODO(lew): We should apply numerics here, so that 'quant' function
    # Can be considered a part of API.
    return self.replace(qvalue=qvalue)  # pytype: disable=attribute-error

  def dequant(self) -> jnp.ndarray:
    """Dequantizes the QTensor."""
    assert self.scale is not None, 'Missing scales when dequantizing a QTensor.'
    msg = (
        'QTensor is manually created without setting a dequant_detype. It can'
        ' be used in dot_general, but to dequantize you need to set its dtype.'
    )
    assert self.dequant_dtype is not None, msg
    assert self.qvalue is not None, _MSG_NO_QVALUE
    ret = self.qvalue
    for scale in self.scale:
      ret = ret.astype(self.dequant_dtype) * scale.astype(self.dequant_dtype)  # pytype: disable=attribute-error
    return ret  # pytype: disable=bad-return-type

  def qvalue_astype(self, dtype) -> Self:
    assert self.qvalue is not None, _MSG_NO_QVALUE
    return self.replace(qvalue=self.qvalue.astype(dtype))  # pytype: disable=attribute-error

  def __getitem__(self, idx: jax_typing.ArrayLike) -> Self:
    """Returns the indexed subtensor on the first axis."""
    assert self.scale_t is None, 'scale_t is not supported in __getitem__'
    assert self.qvalue is not None, _MSG_NO_QVALUE
    qvalue = self.qvalue[idx]
    scale = [s[idx] for s in self.scale]
    return QTensor(
        qvalue=qvalue,
        scale=scale,
        scale_t=self.scale_t,
        dequant_dtype=self.dequant_dtype,
    )

  @property
  def ndim(self) -> int:
    assert self.qvalue is not None, _MSG_NO_QVALUE
    return self.qvalue.ndim  # pytype: disable=attribute-error

  @property
  def shape(self) -> Sequence[int]:
    assert self.qvalue is not None, _MSG_NO_QVALUE
    return self.qvalue.shape  # pytype: disable=attribute-error


def zeros(
    shape: Sequence[int], qdtype: jnp.dtype, dequant_dtype: jnp.dtype
) -> QTensor:
  return QTensor(
      qvalue=jnp.zeros(shape, dtype=qdtype),
      scale=[],
      scale_t=[],
      dequant_dtype=dequant_dtype,
  )


def zeros_with_scale(
    shape: Sequence[int],
    calibration_axis: Sequence[int],
    qdtype: jnp.dtype,
    dequant_dtype: jnp.dtype,
) -> QTensor:
  """Initializes a QTensor with empty qvalue along with empty scale value."""
  scale_shape = list(shape)
  for axis in calibration_axis:
    scale_shape[axis] = 1

  # TODO(lew): hardcode dequant_dtype to bf16. This requires updating
  # other libraries to not break their functionality.
  return QTensor(
      jnp.zeros(shape, dtype=qdtype),
      [jnp.ones(scale_shape, dtype=dequant_dtype)],
      None,
      dequant_dtype=dequant_dtype,
  )


def dynamic_slice(
    operand: QTensor,
    start_indices: Sequence[int],
    slice_sizes: Sequence[int],
) -> QTensor:
  """Dynamically slices the value at start_indices using the given shape."""
  msg = 'scale_t is not supported in the dynamic_slice of a QTensor.'
  assert operand.scale_t is None, msg

  def get_sliced_scales(scale):
    msg = 'Slice sizes must have the same length as operand dims.'
    assert scale.ndim == len(slice_sizes), msg
    scale_start_indices = list(start_indices)
    scale_slice_sizes = list(slice_sizes)
    for axis in range(len(scale_slice_sizes)):
      # slice size must be <= operand shape
      # scale_slice_sizes[dim] = min(scale_slice_sizes[dim], scale.shape[dim])
      if scale.shape[axis] == 1:
        scale_start_indices[axis] = 0
        scale_slice_sizes[axis] = 1
      msg = (
          'We do not support window overflow that is supported in'
          ' jax.lax.dynamic_slices. Please email lew@google.com if you think'
          ' this is wrong.'
      )
      assert (
          scale_start_indices[axis] + scale_slice_sizes[axis]
          <= scale.shape[axis]
      ), msg
    return jax.lax.dynamic_slice(scale, scale_start_indices, scale_slice_sizes)

  return QTensor(
      jax.lax.dynamic_slice(operand.qvalue, start_indices, slice_sizes),
      [get_sliced_scales(s) for s in operand.scale],
      None,
      operand.dequant_dtype,
  )


def dynamic_update_slice(
    operand: QTensor, update: QTensor, start_indices: Sequence[int]
) -> QTensor:
  """Updates the value at start_indices with the given QTensor value."""
  # This function only works for a specific case, i.e., updating an entire slice
  # along the calibration axis.
  msg = 'scale_t is not supported in dynamic_update_slice'
  assert operand.scale_t is None, msg
  op_dd = operand.dequant_dtype
  up_dd = update.dequant_dtype
  assert op_dd == up_dd, f'Dequant dtype mismatch: {op_dd} != {up_dd}'
  ndim = operand.qvalue.ndim
  assert update.qvalue.ndim == ndim
  for scale, update_scale in zip(operand.scale, update.scale):
    assert scale.ndim == ndim
    assert update_scale.ndim == ndim
    for axis in range(ndim):
      if scale.shape[axis] == 1:
        calibration_axis_shape = operand.qvalue.shape[axis]
        update_calibration_axis_shape = update.qvalue.shape[axis]
        msg = (
            'Only updating an entire slice along the calibration axis is'
            f' valid. The calibration axis shape is {calibration_axis_shape}'
            f' but the update slice shape is {update_calibration_axis_shape}.'
        )
        assert update.qvalue.shape[axis] == operand.qvalue.shape[axis], msg
        msg = 'Update scale and scale should be calibrated along the same axis.'
        assert update_scale.shape[axis] == 1, msg
      else:
        # individual scale per element always works
        pass

  qvalues = jax.lax.dynamic_update_slice(
      operand.qvalue, update.qvalue, start_indices
  )
  scales = [
      jax.lax.dynamic_update_slice(scale, update_scale, start_indices)
      for scale, update_scale in zip(operand.scale, update.scale)
  ]

  return QTensor(qvalues, scales, None, operand.dequant_dtype)


def update_frame(operand: QTensor, frame: int, update: QTensor) -> QTensor:
  """Updates the value at frame with the given QTensor value."""
  assert operand.ndim == update.ndim + 1
  assert operand.dequant_dtype == update.dequant_dtype, 'Dequant dtype mismatch'

  return QTensor(
      operand.qvalue.at[frame].set(update.qvalue),
      [
          target_scale.at[frame].set(update_scale)
          for target_scale, update_scale in zip(operand.scale, update.scale)
      ],
      None,
      operand.dequant_dtype,
  )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cost tests for imagenet.models."""

import logging

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized

from aqt.jax_legacy.jax import compute_cost_utils
from aqt.jax_legacy.jax import flax_layers as aqt_flax_layers  # pylint: disable=unused-import
from aqt.jax_legacy.jax import hlo_utils
from aqt.jax_legacy.jax.imagenet import hparams_config
from aqt.jax_legacy.jax.imagenet import models
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4_a2_fixed
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4_a4_fixed
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w8
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w8_a8_fixed
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_bfloat16
from aqt.jax_legacy.jax.imagenet.train_utils import create_model
from aqt.jax_legacy.utils import hparams_utils

from jax import random
import jax.numpy as jnp

FLAGS = flags.FLAGS


class CostTest(parameterized.TestCase):

  def setUp(self):
    super(CostTest, self).setUp()
    self.rng_key = random.PRNGKey(0)

  def _create_hlo_from_resnet_hparams(self, hparams, input_shape):
    """Create an HLO representation from ResNet model and input_shape."""

    # Create model
    rng = random.PRNGKey(0)
    model, init_state = create_model(
        rng,
        input_shape[0],
        input_shape[1],
        jnp.float32,
        hparams.model_hparams,
        train=False)

    # Create HLO
    hlo_proto = hlo_utils.load_hlo_proto_from_model(model, init_state,
                                                    [input_shape])

    del model, init_state
    return hlo_proto

  @parameterized.named_parameters(
      # Unquantized
      dict(
          testcase_name='resnet50_no_quantization',
          base_config_filename=resnet50_bfloat16,
          expected_compute_cost=1046831169536,
          expected_compute_cost_ratio=1.0,
          expected_memory_cost=408046592,
          expected_memory_cost_ratio=1.0,
      ),
      # Weights only
      dict(
          testcase_name='resnet50_full_quantization_weights_only_8',
          base_config_filename=resnet50_w8,
          expected_compute_cost=523415584768,
          expected_compute_cost_ratio=0.5,
          expected_memory_cost=204023296,
          expected_memory_cost_ratio=0.5,
      ),
      dict(
          testcase_name='resnet50_full_quantization_weights_only_4',
          base_config_filename=resnet50_w4,
          expected_compute_cost=261707792384,
          expected_compute_cost_ratio=0.25,
          expected_memory_cost=102011648,
          expected_memory_cost_ratio=0.25,
      ),
      # Weights and activations (fixed)
      dict(
          testcase_name='resnet50_full_quantization_weights_8_fixed_acts_8',
          base_config_filename=resnet50_w8_a8_fixed,
          expected_compute_cost=261707792384,
          expected_compute_cost_ratio=0.25,
          expected_memory_cost=204023296,
          expected_memory_cost_ratio=0.5,
      ),
      dict(
          testcase_name='resnet50_full_quantization_weights_4_fixed_acts_4',
          base_config_filename=resnet50_w4_a4_fixed,
          expected_compute_cost=65426948096,
          expected_compute_cost_ratio=0.0625,
          expected_memory_cost=102011648,
          expected_memory_cost_ratio=0.25,
      ),
      dict(
          testcase_name='resnet50_full_quantization_weights_4_fixed_acts_2',
          base_config_filename=resnet50_w4_a2_fixed,
          expected_compute_cost=32713474048,
          expected_compute_cost_ratio=0.03125,
          expected_memory_cost=102011648,
          expected_memory_cost_ratio=0.25,
      ),
  )  # pylint: disable=line-too-long
  def test_estimate_resnet_cost(self, base_config_filename,
                                expected_compute_cost,
                                expected_compute_cost_ratio,
                                expected_memory_cost,
                                expected_memory_cost_ratio):
    batch_size = 1024
    image_size = 224
    input_channels = 3
    input_shape = (batch_size, image_size, image_size, input_channels)

    logging.info('Testing for %s...', base_config_filename)
    hparams = hparams_utils.load_hparams_from_config_dict(
        hparams_config.TrainingHParams, models.ResNet.HParams,
        base_config_filename.get_config())

    hlo_proto = self._create_hlo_from_resnet_hparams(hparams, input_shape)
    compute_result = compute_cost_utils.estimate_compute_cost(hlo_proto)
    memory_result = compute_cost_utils.estimate_memory_cost(hlo_proto)
    self.assertEqual(compute_result['compute_cost'], expected_compute_cost)
    self.assertAlmostEqual(compute_result['compute_cost_ratio_to_bfloat16'],
                           expected_compute_cost_ratio)
    self.assertEqual(memory_result['memory_cost'], expected_memory_cost)
    self.assertAlmostEqual(memory_result['memory_cost_ratio_to_bfloat16'],
                           expected_memory_cost_ratio)


if __name__ == '__main__':
  FLAGS.metadata_enabled = True  # Passes quantization information to HLO
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2.examples import flax_e2e_model
import aqt.jax.v2.numerics.fp8_numerics as numerics
import jax
import jax.numpy as jnp
import numpy as np


class MyTest(parameterized.TestCase):

  def test_2bit_mantissa(self):
    x = jnp.array(
        [
            [0, 1, -1, 0.1234567, 100, 0.5, 0.00000000001, 0.01],
            [
                -2.9098532,
                -0.8508397,
                -0.14903745,
                -3.572757,
                0.63277215,
                0.7478278,
                1.3707844,
                1.3319879,
            ],
        ],
        dtype=jnp.float32,
    )

    result = numerics.round_to_nearest_even(x, jnp.float8_e5m2)

    np.testing.assert_array_equal(
        result,
        np.array(
            [
                [0, 1, -1, 0.125, 96, 0.5, 0, 10 / 1024],
                [-3, -0.875, -0.15625, -3.5, 0.625, 0.75, 1.25, 1.25],
            ],
            dtype=jnp.float32,
        ),
    )

  def test_3bit_mantissa(self):
    x = jnp.array(
        [
            [0, 1, -1, 0.1234567, 100, 0.5, 0.00000000001, 0.01],
            [
                -2.9098532,
                -0.8508397,
                -0.14903745,
                -3.572757,
                0.63277215,
                0.7478278,
                1.3707844,
                1.3319879,
            ],
        ],
        dtype=jnp.float32,
    )

    result = numerics.round_to_nearest_even(x, jnp.float8_e4m3fn)

    np.testing.assert_array_equal(
        result,
        np.array(
            [
                [0, 1, -1, 0.125, 96, 0.5, 0, 10 / 1024],
                [-3, -0.875, -0.15625, -3.5, 0.625, 0.75, 1.375, 1.375],
            ],
            dtype=jnp.float32,
        ),
    )

  def test_retains_dtype(self):
    result_f16 = numerics.round_to_nearest_even(
        jnp.array([1], dtype=jnp.bfloat16), jnp.float8_e5m2
    )
    result_f32 = numerics.round_to_nearest_even(
        jnp.array([1], dtype=jnp.float32), jnp.float8_e4m3fn
    )

    self.assertEqual(result_f32.dtype, jnp.float8_e4m3fn)
    self.assertEqual(result_f16.dtype, jnp.float8_e5m2)

  @parameterized.parameters([
      dict(fwd_bits="e4m3"),
      dict(fwd_bits="e5m2"),
  ])
  def test_mnist_training(self, fwd_bits: str):
    target_loss = {
        "e4m3": {
            "cpu": [
                # One of them is: milan, rome, haswell
                # Other is: skylake, cascadelake
                3.164344310760498046875000000000,
                3.164344787597656250000000000000,
            ],
            "TPU v2": [3.209973335266113281250000000000],
            "TPU v3": [3.209973335266113281250000000000],
            "TPU v4": [3.210477352142333984375000000000],
            "TPU v5 lite": [3.210477113723754882812500000000],
        },
        "e5m2": {
            "cpu": [
                # One of them is: milan, rome, haswell
                # Other is: skylake, cascadelake
                3.087945222854614257812500000000,
                3.087944984436035156250000000000,
            ],
            "TPU v2": [3.169983625411987304687500000000],
            "TPU v3": [3.169983625411987304687500000000],
            "TPU v4": [3.169979095458984375000000000000],
            "TPU v5 lite": [3.169981479644775390625000000000],
        },
    }

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng = jax.random.split(rng)
    rng, label_rng = jax.random.split(rng)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 8
    ds = {
        "image": jax.random.uniform(key=image_rng, shape=(ds_size, 28, 28, 1)),
        "label": jax.random.randint(
            key=label_rng, shape=(ds_size,), minval=0, maxval=10
        ),
    }

    aqt_cfg = config.config_fwd_fp8(fwd_bits)
    state = flax_e2e_model.create_train_state(init_rng, aqt_cfg)

    _, train_loss, _ = flax_e2e_model.train_epoch(
        state, ds, batch_size=ds_size // 2, rng=input_rng
    )

    device = jax.devices()[0].device_kind
    if train_loss not in target_loss[fwd_bits][device]:
      msg = f"train_loss={train_loss:.30f}, {device=}, {fwd_bits=}"
      self.fail(msg)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for config_schema.py."""

import json

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax.wmt_mlperf.hparams_config_scripts import config_schema


class BaseConfigTest(parameterized.TestCase):

  def test_precision_propagates(self):
    config = config_schema.get_config(
        n_layers=3, use_auto_acts=True, fp_quant=False)

    # Set the global precision to 4 bits.
    config.prec = 4
    # Test that this sets the weight and activation to 4 as well.
    self.assertEqual(config.weight_prec, 4)
    self.assertEqual(config.quant_act.prec, 4)
    # Test that propagates all the way down to the weight precision of layer
    # types and individual layers. As an example of an individual layer, we take
    # the dense1 matmul of the second block of the decoder.
    dense1_block2 = config.model_hparams.decoder.encoder_decoder_1d_blocks[
        1].mlp_block.dense_1
    # Meanwhile, 'dense1' represents the generic configuration of all dense1
    # layers throughout the model.
    dense1 = config.mlp_block.dense_1
    self.assertEqual(dense1.weight_prec, 4)
    self.assertEqual(dense1_block2.weight_prec, 4)

    # Test if we take the same config instance and alter the global precision to
    # 8, it automatically propagates to individual layers.
    config.prec = 8
    self.assertEqual(dense1.weight_prec, 8)
    self.assertEqual(dense1_block2.weight_prec, 8)

    # Test that the precision can be overridden for a specific layer type. We
    # want to verify that the change doesn't back-propagate back to the global
    # precision field but does propagate down to individual layers of that layer
    # type. We only want changes to fields to automatically propagate down the
    # parameter hierarchy, not up.
    dense1.weight_prec = 2
    self.assertEqual(dense1.weight_prec, 2)
    self.assertEqual(dense1_block2.weight_prec, 2)
    self.assertEqual(config.prec, 8)

    # Now update the precision for just a specific layer and check that it
    # doesn't propagate upwards.
    dense1_block2.weight_prec = 1
    self.assertEqual(dense1_block2.weight_prec, 1)
    self.assertEqual(dense1.weight_prec, 2)
    self.assertEqual(config.prec, 8)

  def test_fp_precision_param(self):
    config = config_schema.get_config(
        n_layers=3, use_auto_acts=True, fp_quant=True)
    prec_dict = {
        'is_scaled': False,
        'fp_spec': {
            'exp_min': -3,
            'exp_max': 5,
            'sig_bits': 2
        }
    }
    # Set the global precision to 4 bits.
    config.prec.update(prec_dict)

    # Test that this sets the weight and activation prec as well.
    self.assertEqual(config.weight_prec.to_dict(), prec_dict)
    self.assertEqual(config.quant_act.prec.to_dict(), prec_dict)

    # Test that propagates all the way down to the weight precision of layer
    # types and individual layers. As an example of an individual layer, we take
    # the dense1 matmul of the second block of the decoder.
    dense1_block2 = config.model_hparams.decoder.encoder_decoder_1d_blocks[
        1].mlp_block.dense_1
    # Meanwhile, 'dense1' represents the generic configuration of all dense1
    # layers throughout the model.
    dense1 = config.mlp_block.dense_1
    self.assertEqual(dense1.weight_prec.to_dict(), prec_dict)
    self.assertEqual(dense1_block2.weight_prec.to_dict(), prec_dict)

  @parameterized.parameters(
      dict(n_layers=1), dict(n_layers=3), dict(n_layers=6))
  def test_n_layers_parameter(self, n_layers):
    config = config_schema.get_config(
        n_layers=n_layers, use_auto_acts=True, fp_quant=False)
    self.assertLen(config.model_hparams.encoder.encoder_1d_blocks, n_layers)
    self.assertLen(config.model_hparams.decoder.encoder_decoder_1d_blocks,
                   n_layers)

  def test_auto_acts_parameter(self):
    # If use_auto_acts is False, then the bounds should be a single scalar that
    # specifies the fixed bound; 'None' by default.
    config = config_schema.get_config(
        n_layers=3, use_auto_acts=False, fp_quant=False)
    self.assertIsNone(config.quant_act.bounds)
    # If use_auto_acts is True, it should have the same structure as the
    # GetBounds.Hyper dataclass.
    config = config_schema.get_config(
        n_layers=3, use_auto_acts=True, fp_quant=False)
    self.assertIn('initial_bound', config.quant_act.bounds)

    # Because the config dict is locked, it shouldn't be possible to change it
    # back to fixed bounds if it was created with use_auto_acts=True.
    with self.assertRaises(TypeError):
      config.quant_act.bounds = 1.0

  @parameterized.parameters(
      dict(quantized_reductions=False), dict(quantized_reductions=True))
  def test_softmax_config(self, quantized_reductions):
    base_config = config_schema.get_config(
        n_layers=3, use_auto_acts=True, fp_quant=False)
    softmax_config = config_schema.get_softmax_config(
        quantized=True, quantized_reductions=quantized_reductions)
    softmax_config.quant_hparams.prec.update({
        'exp_min': -3,
        'exp_max': 5,
        'sig_bits': 2
    })
    if quantized_reductions:
      softmax_config.quant_hparams.reduction_prec.update({
          'exp_min': 1,
          'exp_max': 4,
          'sig_bits': 6
      })
    new_config = config_schema.set_global_softmax_config(
        base_config=base_config, softmax_config=softmax_config)
    quant_hparams = new_config.model_hparams.decoder.encoder_decoder_1d_blocks[
        1].enc_dec_attention.attn_acts.softmax.quant_hparams
    self.assertEqual(quant_hparams.prec.to_dict(), {
        'exp_min': -3,
        'exp_max': 5,
        'sig_bits': 2
    })
    if quantized_reductions:
      self.assertEqual(quant_hparams.reduction_prec.to_dict(), {
          'exp_min': 1,
          'exp_max': 4,
          'sig_bits': 6
      })
    else:
      self.assertIsNone(quant_hparams.reduction_prec)

  def test_softmax_raises_with_invalid_quantized_reduction(self):
    with self.assertRaises(ValueError):
      config_schema.get_softmax_config(
          quantized=False, quantized_reductions=True)

  @parameterized.parameters(dict(n_layers=1), dict(n_layers=3))
  def test_schema_matches_expected(self, n_layers):
    # This tests that the schema of the configdict returned by 'config_schema',
    # once all references are resolved, matches an expected schema. 'Schema'
    # here means the names and structure of fields at each level of the
    # configuration hierarchy. A value of 'None' in the expected schemas defined
    # below indicates a real configuration would have a concrete scalar value
    # there.

    sparsity_schema = {
        'type': None,
        'prune_rate': [None, None],  # set to default structured
        'smallest': None,
        'order': None,
        'absolute': None,
        'structure_decay': None,
        'sparse_ste': None
    }

    quant_act_schema = {
        'bounds': {
            'initial_bound': None,
            'stddev_coeff': None,
            'absdev_coeff': None,
            'mix_coeff': None,
            'reset_stats': None,
            'ema_coeff': None,
            'use_cams': None,
            'exclude_zeros': None,
            'use_mean_of_max': None,
            'granularity': None
        },
        'input_distribution': None,
        'prec': None,
        'half_shift': None,
    }

    dense_schema = {
        'weight_prec': None,
        'weight_quant_granularity': None,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'weight_half_shift': None,
        'weight_sparsity': sparsity_schema,
        'act_sparsity': sparsity_schema,
    }

    embedding_schema = {
        'weight_prec': None,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'weight_half_shift': None,
    }

    mlp_block_schema = {
        'dense_1': dense_schema,
        'dense_2': dense_schema,
    }

    fp_schema = {'exp_min': None, 'exp_max': None, 'sig_bits': None}

    layernorm_schema = {
        'quant_hparams': {
            'prec': fp_schema,
            'reduction_prec': fp_schema
        }
    }

    attention_schema = {
        'dense_kqv': dense_schema,
        'dense_out': dense_schema,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'attn_acts': {
            'quant_type': None,
            'attn_act_q': quant_act_schema,
            'attn_act_k': quant_act_schema,
            'attn_act_v': quant_act_schema,
            'attn_act_probs': {
                'bounds': None,
                'input_distribution': None,
                'prec': None,
                'half_shift': None,
            },
        },
    }

    expected_top_level_schema = {
        'metadata': {
            'description': None,
            'hyper_str': None
        },
        'learning_rate_schedule': {
            'factors': None,
            'base_learning_rate': None,
            'warmup_steps': None,
            'decay_factor': None,
            'steps_per_decay': None,
            'steps_per_cycle': None,
        },
        'per_host_batch_size': None,
        'num_train_steps': None,
        'weight_decay': None,
        'beta1': None,
        'beta2': None,
        'eps': None,
        'random_seed': None,
        'hardware_rng': None,
        'activation_bound_update_freq': None,
        'activation_bound_start_step': None,
        'weight_outlier_regularization': None,
        'weight_outlier_regularization_regex': None,
        'prefer_int8_to_int32_dot': None,
        'prec': None,
        'half_shift': None,
        'weight_prec': None,
        'weight_half_shift': None,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'weight_quant_granularity': None,
        'dense': dense_schema,
        'embedding': embedding_schema,
        'mlp_block': mlp_block_schema,
        'attention': attention_schema,
        'sparsity': sparsity_schema,
        'weight_sparsity': sparsity_schema,
        'act_sparsity': sparsity_schema,
        'model_hparams': {
            'emb_dim': None,
            'num_heads': None,
            'qkv_dim': None,
            'mlp_dim': None,
            'share_embeddings': None,
            'logits_via_embedding': None,
            'encoder': {
                'encoder_1d_blocks': [{
                    'mlp_block': mlp_block_schema,
                    'attention': attention_schema,
                    'layer_norm': layernorm_schema
                }] * n_layers,
                'embedding':
                    embedding_schema,
                'layer_norm':
                    layernorm_schema
            },
            'decoder': {
                'encoder_decoder_1d_blocks': [{
                    'mlp_block': mlp_block_schema,
                    'self_attention': attention_schema,
                    'enc_dec_attention': attention_schema,
                    'layer_norm': layernorm_schema
                }] * n_layers,
                'embedding':
                    embedding_schema,
                'logits':
                    dense_schema,
                'layer_norm':
                    layernorm_schema
            },
        },
    }

    config = config_schema.get_config(
        n_layers=n_layers, use_auto_acts=True, fp_quant=False)
    layer_norm_config = config_schema.get_layer_norm_config(
        quantized=True, quantized_reductions=True)
    config = config_schema.set_global_layer_norm_config(config,
                                                        layer_norm_config)
    # This round-trip conversion from JSON forces all references to resolve to
    # concrete values.
    config_reified = json.loads(config.to_json())

    # This test is not interested in checking the specific values of fields in
    # the configuration, but only that the schema of the hierarchies
    # are the same. Thus we all set the value of leaf nodes in the config to
    # 'None' before checking that the actual and expected configuration
    # structures are the same.
    def set_leaves_to_none(config):
      # We are at an intermediate node in the tree-structured input, which could
      # either be in the form of a dictionary or a list of other nodes in the
      # tree.
      if isinstance(config, dict):
        return {key: set_leaves_to_none(value) for key, value in config.items()}
      elif isinstance(config, list):
        return [set_leaves_to_none(value) for value in config]

      # We are at a leaf node in the tree-structured input.
      else:
        return None

    self.assertSameStructure(
        set_leaves_to_none(config_reified), expected_top_level_schema)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.get_bounds."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import test_utils
import flax
from jax import random
import jax.numpy as jnp
import numpy as onp

test_utils.configure_jax()


class GetBoundsTest(parameterized.TestCase):

  def setUp(self):
    super(GetBoundsTest, self).setUp()
    self.rng = random.PRNGKey(0)
    key1, key2 = random.split(self.rng)
    self.key2 = key2
    self.x = random.normal(key1, (4, 3, 2))
    self.x2 = jnp.ones((4, 3, 2))
    self.hyperparam = get_bounds.GetBounds.Hyper(
        initial_bound=6.0,
        stddev_coeff=2.0,
        absdev_coeff=1.5,
        mix_coeff=0.7,
        reset_stats=False,
        granularity=quant_config.QuantGranularity.PER_CHANNEL)

  def init_model(self,
                 update_bounds,
                 update_stats=True,
                 reset_stats=False,
                 use_cams=False,
                 granularity=quant_config.QuantGranularity.PER_TENSOR,
                 ema_coeff=None):
    self.hyperparam = get_bounds.GetBounds.Hyper(
        initial_bound=self.hyperparam.initial_bound,
        stddev_coeff=self.hyperparam.stddev_coeff,
        absdev_coeff=self.hyperparam.absdev_coeff,
        mix_coeff=self.hyperparam.mix_coeff,
        reset_stats=reset_stats,
        use_cams=use_cams,
        ema_coeff=ema_coeff,
        granularity=granularity)
    gb_bounds_params = get_bounds.GetBounds.Params(
        update_bounds=update_bounds, update_stats=update_stats)
    bounds_module = get_bounds.GetBounds(hyper=self.hyperparam)
    init_state = bounds_module.init(
        self.key2, self.x, bounds_params=gb_bounds_params)
    return bounds_module, init_state, gb_bounds_params

  # TODO(shivaniagrawal): parametrize test for different values of axis
  @parameterized.named_parameters(
      dict(testcase_name='update_bound', update_bound=True),
      dict(testcase_name='do_not_update', update_bound=False),
  )
  def test_get_bounds_init(self, update_bound):
    _, init_state, _ = self.init_model(update_bound)
    init_state_stats = init_state['get_bounds']['stats']
    onp.testing.assert_array_equal(init_state_stats.n, 0)
    onp.testing.assert_array_equal(init_state_stats.mean, 0)
    onp.testing.assert_array_equal(init_state_stats.mean_abs, 0)
    onp.testing.assert_array_equal(init_state_stats.mean_sq, 0)
    onp.testing.assert_array_equal(init_state['get_bounds']['bounds'], 6.)

  # TODO(shivaniagrawal): more elaborate testing here as follows:
  # - run with (update_stats, update_bounds) = (False, False)
  # check that neither state changed
  # - run with (update_stats, update_bounds) = (True, False)
  # check that stats.n increased, bound unchanged
  # - run with (update_stats, update_bounds) = (False, True)
  # check that stats.n unchanged but bound updated.
  # - run again with (update_stats, update_bounds) = (False, True)
  # check that both unchanged (update_bounds is idempotent)
  # - run again with (update_stats, update_bounds) = (True, True)
  # check that both changed.

  @parameterized.named_parameters(
      dict(
          testcase_name='update_bound_reset_stats',
          update_bound=True,
          reset_stats=True),
      dict(
          testcase_name='no_update_bound_reset_stats',
          update_bound=False,
          reset_stats=True),
      dict(
          testcase_name='update_bound_no_reset_stats',
          update_bound=True,
          reset_stats=False),
      dict(
          testcase_name='no_update_bound_no_reset_stats',
          update_bound=False,
          reset_stats=False),
  )
  def test_update_stats(self, update_bound, reset_stats):
    model, init_state, params = self.init_model(
        update_bound,
        reset_stats=reset_stats,
        granularity=quant_config.QuantGranularity.PER_TENSOR)
    _, state_0 = model.apply(
        init_state, self.x, bounds_params=params, mutable='get_bounds')

    stats_0_stats = state_0['get_bounds']['stats']
    if reset_stats and update_bound:
      onp.testing.assert_array_equal(stats_0_stats.n, 0)
    else:
      onp.testing.assert_array_equal(stats_0_stats.n, 1)

    _, state = model.apply(
        state_0, self.x2, bounds_params=params, mutable='get_bounds')
    stats = state['get_bounds']['stats']
    if reset_stats and update_bound:
      onp.testing.assert_array_equal(stats.n, 0)
      expected_updated_mean = 0.
      onp.testing.assert_array_equal(expected_updated_mean, stats.mean)
    else:
      onp.testing.assert_array_equal(stats.n, 2)
      expected_updated_mean = 1 / 2 * (1 + (stats_0_stats.mean))
      onp.testing.assert_array_equal(expected_updated_mean, stats.mean)

  @parameterized.named_parameters(
      dict(
          testcase_name='update_bound_reset_stats',
          update_bound=True,
          reset_stats=True),
      dict(
          testcase_name='no_update_bound_reset_stats',
          update_bound=False,
          reset_stats=True),
      dict(
          testcase_name='update_bound_no_reset_stats',
          update_bound=True,
          reset_stats=False),
      dict(
          testcase_name='no_update_bound_no_reset_stats',
          update_bound=False,
          reset_stats=False),
  )
  def test_update_stats_false(self, update_bound, reset_stats):
    model, init_state, params = self.init_model(
        update_bound, update_stats=False, reset_stats=reset_stats)
    _, state_0 = model.apply(
        init_state, self.x, bounds_params=params, mutable='get_bounds')

    stats_0_stats = state_0['get_bounds']['stats']
    onp.testing.assert_array_equal(stats_0_stats.n, 0)

    _, state = model.apply(
        state_0, self.x2, bounds_params=params, mutable='get_bounds')
    onp.testing.assert_array_equal(state['get_bounds']['stats'].n, 0)
    expected_updated_mean = 0.
    onp.testing.assert_array_equal(expected_updated_mean,
                                   state['get_bounds']['stats'].mean)

  @parameterized.named_parameters(
      dict(
          testcase_name='update_bounds_true',
          update_stats=False,
          update_bounds=True),
      dict(
          testcase_name='update_stats_true',
          update_stats=True,
          update_bounds=False),
      dict(testcase_name='both_true', update_stats=True, update_bounds=True),
  )
  def test_update_state_with_mutable_false_context_raises_error(
      self, update_stats, update_bounds):
    model, init_state, _ = self.init_model(True)

    with self.assertRaises(flax.errors.ModifyScopeVariableError):
      model.apply(
          init_state,
          self.x2,
          bounds_params=get_bounds.GetBounds.Params(
              update_stats=update_stats, update_bounds=update_bounds),
          mutable=False)

  @parameterized.named_parameters(
      dict(
          testcase_name='update_bound_no_ucb',
          update_bound=True,
          use_cams=False),
      dict(
          testcase_name='update_bound_with_ucb',
          update_bound=True,
          use_cams=True),
      dict(testcase_name='do_not_update', update_bound=False, use_cams=False),
  )
  def test_get_bounds_update_bounds(self, update_bound, use_cams=False):
    model, init_state, params = self.init_model(update_bound, use_cams=use_cams)
    y, state_0 = model.apply(
        init_state, self.x, bounds_params=params, mutable='get_bounds')
    if not update_bound:
      onp.testing.assert_array_equal(state_0['get_bounds']['bounds'], 6.)
      onp.testing.assert_array_equal(y, 6.)
    else:
      stats_0_stats = state_0['get_bounds']['stats']
      if use_cams:
        expected_y = onp.abs(onp.mean(
            self.x)) + self.hyperparam.stddev_coeff * onp.std(self.x)
      else:
        expected_y = (
            self.hyperparam.stddev_coeff * self.hyperparam.mix_coeff *
            jnp.sqrt(stats_0_stats.mean_sq) + self.hyperparam.absdev_coeff *
            (1 - self.hyperparam.mix_coeff) * stats_0_stats.mean_abs)
      onp.testing.assert_array_equal(state_0['get_bounds']['bounds'], y)
      onp.testing.assert_allclose(expected_y, y)

    y2, state = model.apply(
        state_0, self.x2, bounds_params=params, mutable='get_bounds')
    onp.testing.assert_array_equal(state['get_bounds']['bounds'], y2)

  @parameterized.named_parameters(
      dict(testcase_name='no_ema', ema_coeff=None),
      dict(testcase_name='ema_.8', ema_coeff=0.8),
      dict(testcase_name='ema_.1', ema_coeff=0.1))
  def test_ema_coeff(self, ema_coeff):
    x1 = jnp.array(1.0)
    x2 = jnp.array(-2.0)
    model, state, params = self.init_model(False, ema_coeff=ema_coeff)
    _, state1 = model.apply(
        state, x1, bounds_params=params, mutable='get_bounds')
    _, state2 = model.apply(
        state1, x2, bounds_params=params, mutable='get_bounds')
    stats = state2['get_bounds']['stats']

    def compute_ema_two_steps(x1, x2, alpha):
      initial_value = 0.0
      ema_step_1 = initial_value + alpha * (x1 - initial_value)
      ema_step_2 = ema_step_1 + alpha * (x2 - ema_step_1)
      return ema_step_2

    if ema_coeff is None:
      exp_mean = (x1 + x2) / 2
      exp_mean_sq = (x1**2 + x2**2) / 2
      exp_mean_abs = (jnp.abs(x1) + jnp.abs(x2)) / 2
    else:
      exp_mean = compute_ema_two_steps(x1, x2, ema_coeff)
      exp_mean_sq = compute_ema_two_steps(x1**2, x2**2, ema_coeff)
      exp_mean_abs = compute_ema_two_steps(jnp.abs(x1), jnp.abs(x2), ema_coeff)
    onp.testing.assert_allclose(stats.mean, exp_mean)
    onp.testing.assert_allclose(stats.mean_sq, exp_mean_sq)
    onp.testing.assert_allclose(stats.mean_abs, exp_mean_abs)
    print(stats)
    return

  @parameterized.named_parameters(
      dict(
          testcase_name='no clipping + per channel bounds',
          clipping_coeff=1.0,
          granularity=quant_config.QuantGranularity.PER_CHANNEL),
      dict(
          testcase_name='with clipping + per tensor bounds',
          clipping_coeff=0.5,
          granularity=quant_config.QuantGranularity.PER_TENSOR))
  def test_dynamic_bounds(self, clipping_coeff, granularity):
    hyperparam = get_bounds.DynamicBounds.Hyper(
        granularity=granularity, clipping_coeff=clipping_coeff)
    if granularity == quant_config.QuantGranularity.PER_TENSOR:
      quant_axis = None
    else:
      quant_axis = (0, 1)
    params = get_bounds.DynamicBounds.Params(quant_axis=quant_axis)
    model = get_bounds.DynamicBounds(hyper=hyperparam)
    state = model.init(
        self.key2, self.x, bounds_params=params)

    x = jnp.array([[[2, 3], [5, 4]], [[5, 6], [7, 6]]])
    bounds = model.apply(state, x, bounds_params=params)
    if granularity == quant_config.QuantGranularity.PER_CHANNEL:
      onp.testing.assert_allclose(bounds, jnp.array([[[7.0, 6.0]]]))
    else:
      onp.testing.assert_allclose(bounds, jnp.array(3.5))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mnist example."""

import functools
from typing import Any
from absl import app
from aqt.jax.v2 import config as aqt_config
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2.flax import aqt_flax
from aqt.jax.v2.flax import utils as aqt_flax_utils
from flax import linen as nn
from flax import struct
from flax.metrics import tensorboard
import jax
import jax.numpy as jnp
import numpy as np
import optax
import tensorflow_datasets as tfds


class CNN(nn.Module):
  """A simple CNN model."""
  bn_use_stats: bool
  aqt_cfg: aqt_config.DotGeneral
  quant_mode: aqt_flax_utils.QuantMode = aqt_flax_utils.QuantMode.TRAIN

  @nn.compact
  def __call__(self, x):
    tiling_cfg = tiled_dot_general.Cfg(
        lhs=tiled_dot_general.TensorTiling(
            contraction_axes=[
                tiled_dot_general.AxisTiling(
                    axis=1, tile_count=2, tile_size=None
                ),
            ],
            remaining_axes=[],
        ),
        rhs=tiled_dot_general.TensorTiling(
            contraction_axes=[
                tiled_dot_general.AxisTiling(
                    axis=0, tile_count=2, tile_size=None
                ),
            ],
            remaining_axes=[],
        ),
    )
    aqt_dg = functools.partial(
        aqt_flax.AqtDotGeneral,
        self.aqt_cfg,
        # In nn.Dense, it is RHS that has the kernel.
        rhs_quant_mode=self.quant_mode,
        tiling_cfg=tiling_cfg,
    )
    use_running_avg = not self.bn_use_stats
    x = nn.Conv(features=32, kernel_size=(3, 3))(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = nn.Conv(features=64, kernel_size=(3, 3))(x)
    x = nn.BatchNorm(use_running_average=use_running_avg, dtype=x.dtype)(x)
    x = nn.relu(x)
    x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
    x = x.reshape((x.shape[0], -1))  # flatten
    x = nn.Dense(features=256, dot_general_cls=aqt_dg)(x)
    x = nn.relu(x)
    x = nn.Dense(features=10, dot_general_cls=aqt_dg)(x)

    # Simple demonstration of how to quantize einsum.
    identity = jnp.identity(10, dtype=x.dtype)
    einsum = aqt_flax.AqtEinsum(self.aqt_cfg, lhs_quant_mode=self.quant_mode)
    # Note for AQT developers:
    #   This equation is harder because jnp.einsum and einsum swap lhs and rhs.
    x = einsum('bc,ab->ac', identity, x)
    return x


@functools.partial(jax.jit, static_argnums=(3,))
def apply_model(state, images, labels, train):
  """Computes gradients, loss and accuracy for a single batch."""

  cnn = state.cnn_train if train else state.cnn_eval
  def loss_fn(model):
    logits, updated_var = cnn.apply(
        model,
        images,
        rngs={'params': jax.random.PRNGKey(0)},
        mutable=True,
    )
    one_hot = jax.nn.one_hot(labels, 10)
    loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
    return loss, (logits, updated_var)

  grad_fn = jax.value_and_grad(loss_fn, has_aux=True, allow_int=True)
  aux, grads = grad_fn(state.model)
  loss, (logits, updated_var) = aux
  accuracy = jnp.mean(jnp.argmax(logits, -1) == labels)
  return grads, loss, accuracy, updated_var


@jax.jit
def update_model(state, grads, updated_var):
  params = state.model['params']
  param_grad = grads['params']
  updates, new_opt_state = state.tx.update(param_grad, state.opt_state, params)
  new_params = optax.apply_updates(params, updates)
  updated_var.update(params=new_params)
  return state.replace(
      model=updated_var,
      opt_state=new_opt_state,
  )


def train_epoch(state, train_ds, batch_size, rng):
  """Train for a single epoch."""
  train_ds_size = len(train_ds['image'])
  steps_per_epoch = train_ds_size // batch_size

  perms = jax.random.permutation(rng, len(train_ds['image']))
  perms = perms[: steps_per_epoch * batch_size]  # skip incomplete batch
  perms = perms.reshape((steps_per_epoch, batch_size))

  epoch_loss = []
  epoch_accuracy = []

  for perm in perms:
    batch_images = train_ds['image'][perm, ...]
    batch_labels = train_ds['label'][perm, ...]
    grads, loss, accuracy, updated_var = apply_model(
        state, batch_images, batch_labels, train=True
    )
    state = update_model(state, grads, updated_var)
    epoch_loss.append(loss)
    epoch_accuracy.append(accuracy)
  train_loss = np.mean(epoch_loss)
  train_accuracy = np.mean(epoch_accuracy)
  return state, train_loss, train_accuracy


def get_datasets():
  """Load MNIST train and test datasets into memory."""
  print('get_datasets started')
  ds_builder = tfds.builder('mnist')
  ds_builder.download_and_prepare()
  train_ds = tfds.as_numpy(ds_builder.as_dataset(split='train', batch_size=-1))
  test_ds = tfds.as_numpy(ds_builder.as_dataset(split='test', batch_size=-1))
  train_ds['image'] = jnp.float32(train_ds['image']) / 255.0
  test_ds['image'] = jnp.float32(test_ds['image']) / 255.0
  print('get_datasets DONE')
  return train_ds, test_ds


class TrainState(struct.PyTreeNode):
  """Train state."""

  cnn_train: Any = struct.field(pytree_node=False)
  cnn_eval: Any = struct.field(pytree_node=False)
  model: Any = struct.field(pytree_node=True)
  tx: optax.GradientTransformation = struct.field(pytree_node=False)
  opt_state: optax.OptState = struct.field(pytree_node=True)


def create_train_state(rng, aqt_cfg):
  """Creates initial `TrainState`."""
  cnn_train = CNN(bn_use_stats=True, aqt_cfg=aqt_cfg)
  model = cnn_train.init({'params': rng}, jnp.ones([1, 28, 28, 1]))
  learning_rate = 0.1
  momentum = 0.9
  tx = optax.sgd(learning_rate, momentum)
  cnn_eval = CNN(bn_use_stats=False, aqt_cfg=aqt_cfg)
  return TrainState(
      cnn_train=cnn_train,
      cnn_eval=cnn_eval,
      model=model,
      tx=tx,
      opt_state=tx.init(model['params']),
  )


def train_and_evaluate(
    num_epochs: int, workdir: str, aqt_cfg: aqt_config.DotGeneral
) -> TrainState:
  """Execute model training and evaluation loop."""
  train_ds, test_ds = get_datasets()
  rng = jax.random.key(0)

  summary_writer = tensorboard.SummaryWriter(workdir)

  rng, init_rng = jax.random.split(rng)
  state = create_train_state(init_rng, aqt_cfg)

  batch_size = 128
  for epoch in range(1, num_epochs + 1):
    rng, input_rng = jax.random.split(rng)
    state, train_loss, train_accuracy = train_epoch(
        state, train_ds, batch_size, input_rng
    )
    _, test_loss, test_accuracy, _ = apply_model(
        state, test_ds['image'], test_ds['label'], train=False
    )

    print(
        'epoch:% 3d, train_loss: %.30f, train_accuracy: %.30f, test_loss:'
        ' %.30f, test_accuracy: %.30f'
        % (
            epoch,
            train_loss,
            train_accuracy * 100,
            test_loss,
            test_accuracy * 100,
        ),
        flush=True,
    )

    summary_writer.scalar('train_loss', train_loss, epoch)
    summary_writer.scalar('train_accuracy', train_accuracy, epoch)
    summary_writer.scalar('test_loss', test_loss, epoch)
    summary_writer.scalar('test_accuracy', test_accuracy, epoch)

  summary_writer.flush()
  return state


def serving_conversion(train_state):
  """Model conversion (quantized weights freezing)."""
  aqt_cfg = train_state.cnn_eval.aqt_cfg
  input_shape = (1, 28, 28, 1)
  cnn_freeze = CNN(
      bn_use_stats=False,
      aqt_cfg=aqt_cfg,
      quant_mode=aqt_flax_utils.QuantMode.CONVERT,
  )
  _, model_serving = cnn_freeze.apply(
      train_state.model,
      jnp.ones(input_shape),
      rngs={'params': jax.random.PRNGKey(0)},
      mutable=True,
  )
  cnn_serve = CNN(
      bn_use_stats=False,
      aqt_cfg=aqt_cfg,
      quant_mode=aqt_flax_utils.QuantMode.SERVE,
  )

  return cnn_serve.apply, model_serving


@jax.jit
def serve(state):
  """Take train state, freeze integer weights, and serve."""
  # get sample serving data
  _, test_ds = get_datasets()
  sample_image, sample_label = test_ds['image'][:64], test_ds['label'][:64]
  # serving
  serve_fn, model_serving = serving_conversion(state)
  logits = serve_fn(
      model_serving, sample_image, rngs={'params': jax.random.PRNGKey(0)}
  )
  # compute serving loss
  one_hot = jax.nn.one_hot(sample_label, 10)
  loss = jnp.mean(optax.softmax_cross_entropy(logits=logits, labels=one_hot))
  return loss


def serve_fn_hlo(state):
  """Example on how to inspect HLO to verify if everything was quantized."""
  # get sample serving data
  _, test_ds = get_datasets()
  sample_image = test_ds['image'][:64]
  # serving
  serve_fn, model_serving = serving_conversion(state)
  # The following XLA graph is only needed for debugging purpose
  hlo = jax.xla_computation(serve_fn)(
      model_serving,
      sample_image,
      rngs={'params': jax.random.PRNGKey(0)},
  ).as_hlo_module()
  return hlo


def main(argv):
  del argv
  aqt_cfg = aqt_config.fully_quantized(fwd_bits=8, bwd_bits=8)
  state = train_and_evaluate(
      num_epochs=2, workdir='/tmp/aqt_mnist_example', aqt_cfg=aqt_cfg
  )
  loss = serve(state)
  print('serve loss on sample ds: {}'.format(loss))


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.train_utils."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import train_utils


class UpdateUtilsTest(parameterized.TestCase):

  @parameterized.parameters(
      # No updates before the start step of 5
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 0,
          'should_update': False
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 3,
          'should_update': False
      },

      # Updates expected every 10 steps after step 5 and no time else
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 5,
          'should_update': True
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 15,
          'should_update': True
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 25,
          'should_update': True
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 13,
          'should_update': False
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 18,
          'should_update': False
      },

      # Test an update frequency of -1, which indicates no bounds update
      {
          'frequency': -1,
          'start_step': 5,
          'current_step': 5,
          'should_update': True
      },
      {
          'frequency': -1,
          'start_step': 5,
          'current_step': 6,
          'should_update': False
      },
      {
          'frequency': -1,
          'start_step': 5,
          'current_step': 4,
          'should_update': False
      },

      # Test a start step of -1, which indicates no bounds update at any step.
      {
          'frequency': -1,
          'start_step': -1,
          'current_step': 3,
          'should_update': False
      },
      {
          'frequency': 5,
          'start_step': -1,
          'current_step': 12,
          'should_update': False
      })
  def test_should_update_bounds(self, frequency, start_step, current_step,
                                should_update):
    self.assertEqual(
        train_utils.should_update_bounds(
            activation_bound_update_freq=frequency,
            activation_bound_start_step=start_step,
            step=current_step), should_update)

  @parameterized.parameters(
      # No updates before the start step of 5
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 0,
          'should_update': False
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 3,
          'should_update': False
      },

      # Updates expected every 10 steps after step 5 and no time else
      {
          'frequency': 8,
          'start_step': 5,
          'current_step': 5,
          'should_update': True
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 15,
          'should_update': True
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 25,
          'should_update': True
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 13,
          'should_update': False
      },
      {
          'frequency': 10,
          'start_step': 5,
          'current_step': 18,
          'should_update': False
      },

      # Test frequency of 0, which indicates no update except at start step
      {
          'frequency': 0,
          'start_step': 5,
          'current_step': 5,
          'should_update': True
      },
      {
          'frequency': 0,
          'start_step': 5,
          'current_step': 10,
          'should_update': False
      },
      {
          'frequency': 0,
          'start_step': 5,
          'current_step': 4,
          'should_update': False
      },

      # Test a start step of -1, which indicates no mask update at any step.
      {
          'frequency': 0,
          'start_step': -1,
          'current_step': 3,
          'should_update': False
      },
      {
          'frequency': 5,
          'start_step': -1,
          'current_step': 5,
          'should_update': False
      })
  def test_should_update_sparsity_mask(self, frequency, start_step,
                                       current_step, should_update):
    self.assertEqual(
        train_utils.update_sparsity_mask(
            sparsity_start_step=start_step,
            sparsity_update_freq=frequency,
            step=current_step), should_update)

  @parameterized.parameters(
      {
          'step': 70,
          'sparsity_start_step': 10,
          'sparsity_update_freq': 10,
          'num_update_sparsity': 6,
      },
      {
          'step': 100,
          'sparsity_start_step': 10,
          'sparsity_update_freq': 10,
          'num_update_sparsity': 8,
      },
      {
          'step': 10,
          'sparsity_start_step': 10,
          'sparsity_update_freq': 10,
          'num_update_sparsity': 0,
      },
      {
          'step': 100,
          'sparsity_start_step': 10,
          'sparsity_update_freq': 0,
          'num_update_sparsity': 0,
      },)
  def test_num_update_sparsity(self, step, sparsity_start_step,
                               sparsity_update_freq, num_update_sparsity):
    dc = train_utils.get_dynamic_context_for_step(
        activation_bound_update_freq=1,
        activation_bound_start_step=1,
        step=step,
        collect_acts_stats=False,
        prefer_int8_to_int32_dot=False,
        sparsity_start_step=sparsity_start_step,
        sparsity_update_freq=sparsity_update_freq)
    self.assertEqual(dc.num_update_sparsity, num_update_sparsity)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.primitives."""
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import primitives
from aqt.jax_legacy.jax import test_utils
import jax
import jax.numpy as jnp
import numpy as onp

fp32 = onp.float32
test_utils.configure_jax()


class PrimitivesTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name='floor_clip_uint1',
          prec=1,
      ),
      dict(
          testcase_name='floor_clip_uint4',
          prec=4,
      ),
      dict(
          testcase_name='floor_clip_uint8',
          prec=8,
      ),
      dict(
          testcase_name='floor_clip_uint16',
          prec=16,
      ),
  )
  def test_floor_and_clip_to_unsigned_int(self, prec):
    x = jnp.array(fp32(2.0**5 * onp.random.uniform(0, 1.0, size=(1024, 1))))
    y = primitives.floor_and_clip_to_unsigned_int(
        x, prec=prec, dtype=x.dtype, half_shift=False)
    self.assertGreaterEqual(onp.min(y), 0.0)
    self.assertLessEqual(onp.max(y), fp32(2**prec - 1))
    onp.testing.assert_allclose(y, onp.around(y))

  @parameterized.named_parameters(
      dict(testcase_name='round_clip_sint1', prec=1),
      dict(testcase_name='round_clip_sint4', prec=4),
      dict(testcase_name='round_clip_sint8', prec=8),
      dict(testcase_name='round_clip_sint16', prec=16))
  def test_round_and_clip_to_signed_int(self, prec):
    np_x = fp32(2.0**5 * onp.random.uniform(-1.0, 1.0, size=(1024, 1)))
    x = jnp.array(np_x)
    y = primitives.round_and_clip_to_signed_int(
        x, prec=prec, dtype=x.dtype, half_shift=False)
    ubound = fp32(2.0**(prec - 1) - 1)
    lbound = fp32(-2.0**(prec - 1) + 1)
    self.assertGreaterEqual(onp.min(y), lbound)
    self.assertLessEqual(onp.max(y), ubound)
    onp.testing.assert_allclose(y, onp.around(y))
    onp.testing.assert_allclose(
        y, onp.clip(onp.floor(np_x + 0.5), a_min=lbound, a_max=ubound))

  @parameterized.named_parameters(
      dict(testcase_name='round_clip_sint1_half_shift', prec=1),
      dict(testcase_name='round_clip_sint4_half_shift', prec=4),
      dict(testcase_name='round_clip_sint8_half_shift', prec=8),
      dict(testcase_name='round_clip_sint16_half_shift', prec=16))
  def test_round_and_clip_to_signed_int_bound_half_shift(self, prec):
    np_x = fp32(2.0**5 * onp.random.uniform(-1.0, 1.0, size=(1024, 1)))
    x = jnp.array(np_x)
    y = primitives.round_and_clip_to_signed_int(
        x, prec=prec, dtype=x.dtype, half_shift=True)
    ubound = fp32(2.0**(prec - 1))
    lbound = fp32(-2.0**(prec - 1))
    self.assertGreaterEqual(onp.min(y), lbound)
    self.assertLessEqual(onp.max(y), ubound)
    onp.testing.assert_allclose(y, onp.floor(y) + 0.5)
    epsilon = 2**(-7)  # epsilon prevents flooring values to 1 (expect -1/0)
    onp.testing.assert_allclose(
        y,
        onp.floor(
            onp.clip(np_x, a_min=lbound + epsilon, a_max=ubound - epsilon)) +
        0.5)

  def test_round_and_clip_to_signed_int_half_shift(self):
    # centers_2bit = [-1.5, -0.5, 0.5, 1.5]
    q_inp = [-2.1, -1.9, -1.1, -0.9, -0.1, +0.1, +0.9, +1.1, +1.9, +2.1]
    q2bit = [-1.5, -1.5, -1.5, -0.5, -0.5, +0.5, +0.5, +1.5, +1.5, +1.5]
    q1bit = [-0.5, -0.5, -0.5, -0.5, -0.5, +0.5, +0.5, +0.5, +0.5, +0.5]
    q2gra = [+0.0, +1.0, +1.0, +1.0, +1.0, +1.0, +1.0, +1.0, +1.0, +0.0]
    q1gra = [+0.0, +0.0, +0.0, +1.0, +1.0, +1.0, +1.0, +0.0, +0.0, +0.0]
    q_inp = jnp.array(q_inp)
    q2bit = jnp.array(q2bit)
    q1bit = jnp.array(q1bit)
    q2gra = jnp.array(q2gra)
    q1gra = jnp.array(q1gra)

    def quantize1bit(x):
      return primitives.round_and_clip_to_signed_int(
          x, prec=1, half_shift=True, dtype=jnp.float32)

    def quantize2bit(x):
      return primitives.round_and_clip_to_signed_int(
          x, prec=2, half_shift=True, dtype=jnp.float32)

    quantize1bit_grad = jax.vmap(jax.grad(quantize1bit))
    quantize2bit_grad = jax.vmap(jax.grad(quantize2bit))

    # TODO(lew): Resolve the error and use self.assertEqual instead.
    # ValueError: The truth value of an array with more than one element is
    # ambiguous. Use a.any() or a.all()
    def assert_equal(a, b):
      self.assertEqual(jnp.sum(jnp.abs(a - b)), 0.0)

    assert_equal(quantize1bit(q_inp), q1bit)
    assert_equal(quantize2bit(q_inp), q2bit)
    assert_equal(quantize1bit_grad(q_inp), q1gra)
    assert_equal(quantize2bit_grad(q_inp), q2gra)

  @parameterized.named_parameters(
      dict(testcase_name='round', op_with_ste=primitives.round_with_gradient),
      dict(testcase_name='floor', op_with_ste=primitives.floor_with_gradient))
  def test_ste(self, op_with_ste):
    x = jnp.array([1.2, 3.8])
    op_grad = jax.grad(lambda x: jnp.sum(op_with_ste(x)))
    onp.testing.assert_array_equal(op_grad(x), 1.0)

  def test_grad_of_round_and_clip_to_signed_int(self):
    x = jnp.array([-1000.5, 5.2, 130.7])
    # Since -1000.5 and 900.3 are outside the 8-bit signed clipping range of
    # [-127, 127], we expect their gradient to be 0. Since the rounding and
    # int-casting use the STE, we expect the gradient of 5.2 to be 1.
    @jax.grad
    def grad_fn(x):
      return jnp.sum(
          primitives.round_and_clip_to_signed_int(
              x, prec=8, dtype=x.dtype, half_shift=False))

    onp.testing.assert_array_equal(grad_fn(x), [0.0, 1.0, 0.0])

  def test_grad_of_round_and_clip_to_unsigned_int(self):
    x = jnp.array([-1.5, 5.2, 300.2])
    # Since -1.5 and 300.2 are outside the 8-bit unsigned clipping range of [0,
    # 255], we expect their gradient to be 0. Since the rounding and int-casting
    # use the STE, we expect the gradient of 5.2 to be 1.
    @jax.grad
    def grad_fn(x):
      return jnp.sum(
          primitives.floor_and_clip_to_unsigned_int(
              x, prec=8, dtype=x.dtype, half_shift=False))

    onp.testing.assert_array_equal(grad_fn(x), [0.0, 1.0, 0.0])


if __name__ == '__main__':
  absltest.main()
